import datetime as dt
from decimal import Decimal as D

from blazeweb.tasks import attributes
from sqlalchemybwc import db

from webgrid_blazeweb_ta.model.orm import Email, Person, Status


@attributes('~dev')
def action_40_base_data():
    stat_open = Status.add_iu(label='open')
    stat_pending = Status.add_iu(label='pending')
    stat_closed = Status.add_iu(label='closed', flag_closed=1)

    for x in range(1, 50):
        p = Person()
        p.firstname = f'fn{x:03d}'
        p.lastname = f'ln{x:03d}'
        p.sortorder = x
        p.numericcol = D('29.26') * x / D('.9')
        if x < 90:
            p.createdts = dt.datetime.now()
        db.sess.add(p)
        p.emails.append(Email(email=f'email{x:03d}@example.com'))
        p.emails.append(Email(email=f'email{x:03d}@gmail.com'))
        if x % 4 == 1:
            p.status = stat_open
        elif x % 4 == 2:
            p.status = stat_pending
        elif x % 4 == 0:
            p.status = None
        else:
            p.status = stat_closed

    db.sess.commit()



from blazeweb.content import getcontent
from blazeweb.globals import ag, rg, user
from blazeweb.routing import abs_static_url
from blazeweb.templating.jinja import content_filter
from blazeweb.utils import abort
from blazeweb.wrappers import StreamResponse
from jinja2.exceptions import TemplateNotFound
from sqlalchemybwc import db as sabwc_db

from webgrid import BaseGrid
from webgrid.extensions import FrameworkManager, gettext
from webgrid.renderers import render_html_attributes


class WebGrid(FrameworkManager):
    def __init__(self, db=None, component='webgrid'):
        super().__init__(db=db or sabwc_db)
        self.component = component

    def init(self):
        ag.tplengine.env.filters['wg_safe'] = content_filter
        ag.tplengine.env.filters['wg_attributes'] = render_html_attributes
        ag.tplengine.env.filters['wg_gettext'] = gettext

    def sa_query(self, *args, **kwargs):
        return self.db.sess.query(*args, **kwargs)

    def request_form_args(self):
        """Return POST request args."""
        return rg.request.form

    def request_json(self):
        """Return json body of request."""
        return rg.request.json

    def request_url_args(self):
        """Return GET request args."""
        return rg.request.args

    def csrf_token(self):
        raise NotImplementedError

    def web_session(self):
        return user

    def persist_web_session(self):
        pass

    def flash_message(self, category, message):
        user.add_message(category, message)

    def request(self):
        return rg.request

    def static_url(self, url_tail):
        return abs_static_url(f'component/webgrid/{url_tail}')

    def file_as_response(self, data_stream, file_name, mime_type):
        rp = StreamResponse(data_stream)
        rp.headers['Content-Type'] = mime_type
        if file_name is not None:
            rp.headers['Content-Disposition'] = f'attachment; filename={file_name}'
        abort(rp)

    def render_template(self, endpoint, **kwargs):
        try:
            return getcontent(endpoint, **kwargs)
        except TemplateNotFound:
            if ':' in endpoint:
                raise
            return getcontent(f'{self.component}:{endpoint}', **kwargs)


wg_blaze_manager = WebGrid()


class Grid(BaseGrid):
    manager = wg_blaze_manager



from typing import ClassVar
from unittest import mock

import flask
import flask_webtest
import flask_wtf
import pytest

from webgrid import BaseGrid, Column
from webgrid.flask import WebGrid, WebGridAPI
from webgrid.renderers import JSON, Renderer


@pytest.fixture
def app():
    app = flask.Flask(__name__)
    app.secret_key = 'only-testing-api'
    app.config['TESTING'] = True
    yield app


@pytest.fixture
def csrf(app):
    csrf = flask_wtf.CSRFProtect()
    csrf.init_app(app)
    yield csrf


@pytest.fixture
def test_app(app):
    yield flask_webtest.TestApp(app)


@pytest.fixture
def api_manager(app):
    manager = WebGridAPI()
    manager.init_app(app)
    yield manager


@pytest.fixture
def api_manager_with_csrf(csrf, app):
    """Technically, this is the same as having ``csrf, api_manager`` as fixtures on
    the test method and turning on csrf_protection. But the order matters, so for
    consistent results this separate combined fixture is provided."""
    manager = WebGridAPI()
    manager.csrf_protection = True
    manager.init_app(app)
    yield manager


class DummyMixin:
    Column('Foo', 'foo')

    @property
    def records(self):
        return [{'foo': 'bar'}]

    @property
    def record_count(self):
        return 1


def create_grid_cls(grid_manager):
    class Grid(DummyMixin, BaseGrid):
        manager = grid_manager
        allowed_export_targets: ClassVar = {'json': JSON}

    return Grid


def register_grid(manager, identifier, grid_cls):
    manager.register_grid(identifier, grid_cls)


class TestFlaskAPI:
    def post_data(self, **kwargs):
        data = {
            'search_expr': '',
            'filters': {},
            'sort': [],
            'paging': {'per_page': 50, 'on_page': 1},
            'export_to': None,
        }
        data.update(**kwargs)
        return data

    def test_default_route(self, api_manager, test_app):
        register_grid(api_manager, 'foo', create_grid_cls(api_manager))
        resp = test_app.post_json('/webgrid-api/foo', {})
        assert resp.json['records'] == [{'foo': 'bar'}]

    def test_default_route_count(self, api_manager, test_app):
        register_grid(api_manager, 'foo', create_grid_cls(api_manager))
        resp = test_app.post_json('/webgrid-api/foo/count', {})
        assert resp.json['count'] == 1

    def test_custom_route(self, app, test_app):
        class GridManager(WebGridAPI):
            api_route = '/custom-routing/<grid_ident>'

        api_manager = GridManager()
        api_manager.init_app(app)
        register_grid(api_manager, 'foo', create_grid_cls(api_manager))
        resp = test_app.post_json('/custom-routing/foo', {})
        assert resp.json['records'] == [{'foo': 'bar'}]

    def test_grid_not_registered(self, api_manager, test_app):
        test_app.post_json('/webgrid-api/foo', {}, status=404)

    def test_grid_registered_twice(self, api_manager, test_app):
        register_grid(api_manager, 'foo', create_grid_cls(api_manager))
        with pytest.raises(Exception, match='API grid_ident must be unique'):
            register_grid(api_manager, 'foo', create_grid_cls(api_manager))

    def test_csrf_missing_token(self, api_manager_with_csrf, test_app):
        register_grid(api_manager_with_csrf, 'foo', create_grid_cls(api_manager_with_csrf))
        resp = test_app.post_json('/webgrid-api/foo', {}, status=400)
        assert 'The CSRF token is missing.' in resp

    def test_csrf_missing_token_but_no_protection(self, api_manager, test_app):
        register_grid(api_manager, 'foo', create_grid_cls(api_manager))
        test_app.post_json('/webgrid-api/foo', {}, status=200)

    def test_csrf_missing_session(self, api_manager_with_csrf, test_app):
        register_grid(api_manager_with_csrf, 'foo', create_grid_cls(api_manager_with_csrf))
        with test_app.app.test_request_context():
            csrf_token = flask_wtf.csrf.generate_csrf()
        resp = test_app.post_json(
            '/webgrid-api/foo',
            {},
            headers={'X-CSRFToken': csrf_token},
            status=400,
        )
        assert 'The CSRF session token is missing.' in resp

    def test_csrf_invalid(self, api_manager_with_csrf, test_app):
        register_grid(api_manager_with_csrf, 'foo', create_grid_cls(api_manager_with_csrf))
        test_app.get('/webgrid-api/testing/__csrf__')
        resp = test_app.post_json(
            '/webgrid-api/foo',
            {},
            headers={'X-CSRFToken': 'my-bad-token'},
            status=400,
        )
        assert 'The CSRF token is invalid.' in resp

    def test_csrf_protected(self, api_manager_with_csrf, test_app):
        register_grid(api_manager_with_csrf, 'foo', create_grid_cls(api_manager_with_csrf))
        csrf_token = test_app.get('/webgrid-api/testing/__csrf__').body
        resp = test_app.post_json('/webgrid-api/foo', {}, headers={'X-CSRFToken': csrf_token})
        assert resp.json['records'] == [{'foo': 'bar'}]

    def test_csrf_token_route_only_testing(self, app, csrf, test_app):
        manager = WebGridAPI()
        with mock.patch.dict(app.config, {'TESTING': False}):
            manager.init_app(app)

        test_app.get('/webgrid-api/testing/__csrf__', status=404)

    def test_grid_has_renderer(self, api_manager, test_app):
        class MyRenderer(Renderer):
            name = 'test'

            def render(self):
                return '"render result"'

        class Grid(DummyMixin, BaseGrid):
            manager = api_manager
            allowed_export_targets: ClassVar = {'json': MyRenderer}

        register_grid(api_manager, 'foo', Grid)
        resp = test_app.post_json('/webgrid-api/foo', {})
        assert resp.json == 'render result'

    def test_grid_default_renderer(self, api_manager, test_app):
        class Grid(DummyMixin, BaseGrid):
            manager = api_manager

        register_grid(api_manager, 'foo', Grid)
        resp = test_app.post_json('/webgrid-api/foo', {})
        assert resp.json['records'] == [{'foo': 'bar'}]

    def test_grid_auth(self, api_manager, test_app):
        class Grid(DummyMixin, BaseGrid):
            manager = api_manager

            def check_auth(self):
                if b'bad' in flask.request.query_string:
                    flask.abort(403)

        register_grid(api_manager, 'foo', Grid)
        resp = test_app.post_json('/webgrid-api/foo', {})
        assert resp.json['records'] == [{'foo': 'bar'}]

        test_app.post_json('/webgrid-api/foo?bad', {}, status=403)

    def test_grid_args_applied(self, api_manager, test_app):
        Grid = create_grid_cls(api_manager)
        Grid.record_count = 1
        register_grid(api_manager, 'foo', Grid)
        post_data = self.post_data(sort=[{'key': 'foo', 'flag_desc': True}])
        resp = test_app.post_json('/webgrid-api/foo', post_data)
        assert resp.json['settings']['sort'] == [{'key': 'foo', 'flag_desc': True}]

    def test_grid_args_applied_alt_manager(self, api_manager, test_app):
        Grid = create_grid_cls(WebGrid())
        Grid.record_count = 1
        register_grid(api_manager, 'foo', Grid)
        post_data = self.post_data(sort=[{'key': 'foo', 'flag_desc': True}])
        resp = test_app.post_json('/webgrid-api/foo', post_data)
        assert resp.json['settings']['sort'] == [{'key': 'foo', 'flag_desc': True}]

    def test_grid_auth_precedes_args(self, api_manager, test_app):
        class Grid(DummyMixin, BaseGrid):
            manager = api_manager

            def check_auth(self):
                flask.abort(403)

        register_grid(api_manager, 'foo', Grid)
        with mock.patch.object(api_manager, 'get_args', autospec=True, spec_set=True) as m_args:
            test_app.post_json('/webgrid-api/foo', {}, status=403)

        assert not m_args.call_count

    def test_grid_export(self, api_manager, test_app):
        class Grid(DummyMixin, BaseGrid):
            manager = api_manager

        register_grid(api_manager, 'foo', Grid)
        post_data = self.post_data(export_to='xlsx')
        resp = test_app.post_json('/webgrid-api/foo', post_data)
        assert 'spreadsheetml' in resp.headers['Content-Type']

    def test_grid_export_limit_exceeded(self, api_manager, test_app):
        class Grid(DummyMixin, BaseGrid):
            manager = api_manager

            @property
            def record_count(self):
                return 2000000

        register_grid(api_manager, 'foo', Grid)
        post_data = self.post_data(export_to='xlsx')
        resp = test_app.post_json('/webgrid-api/foo', post_data)
        assert resp.json['error'] == 'too many records for render target'



import re

import pytest

import webgrid.types as types


class TestGridSettings:
    def ok_values(self, **kwargs):
        return {
            'search_expr': 'foo',
            'filters': {
                'test': {'op': 'eq', 'value1': 'toast', 'value2': 'taft'},
                'test2': {'op': 'in', 'value1': 'tarp', 'value2': None},
            },
            'paging': {'pager_on': True, 'on_page': 2, 'per_page': 20},
            'sort': [{'key': 'bar', 'flag_desc': False}, {'key': 'baz', 'flag_desc': True}],
            **kwargs,
        }

    def test_from_dict(self):
        data = self.ok_values()
        assert types.GridSettings.from_dict(data) == types.GridSettings(
            search_expr='foo',
            filters={
                'test': types.Filter(op='eq', value1='toast', value2='taft'),
                'test2': types.Filter(op='in', value1='tarp'),
            },
            paging=types.Paging(pager_on=True, on_page=2, per_page=20),
            sort=[types.Sort(key='bar', flag_desc=False), types.Sort(key='baz', flag_desc=True)],
            export_to=None,
        )

    def test_from_dict_missing_keys(self):
        assert types.GridSettings.from_dict({}) == types.GridSettings(
            search_expr=None,
            filters={},
            paging=types.Paging(pager_on=False, on_page=None, per_page=None),
            sort=[],
            export_to=None,
        )

    def test_from_dict_invalid_values(self):
        data = self.ok_values(paging={'per_page': 'foo'})
        with pytest.raises(
            types.ValidationError,
            match='Received per_page=foo; should be of type int',
        ):
            types.GridSettings.from_dict(data)

    @pytest.mark.parametrize(
        'subobject,input_data',
        (
            ('Sort', {'sort': [{'key': 'foo', 'flag_desc': False, 'bar': 'baz'}]}),
            ('Paging', {'paging': {'bar': 'baz'}}),
            ('Filter', {'filters': {'test': {'op': 'eq', 'value1': 'foo', 'bar': 'baz'}}}),
        ),
    )
    def test_from_dict_extra_values(self, subobject, input_data):
        data = self.ok_values(**input_data)
        with pytest.raises(
            types.ValidationError,
            match=re.compile(
                f"{subobject}:.*__init__\\(\\) got an unexpected keyword argument 'bar'",
            ),
        ):
            types.GridSettings.from_dict(data)

    def test_from_dict_missing_values(self):
        data = self.ok_values(sort=[{'flag_desc': False}])
        with pytest.raises(
            types.ValidationError,
            match=re.compile(r"Sort:.*__init__\(\) missing 1 required positional argument: 'key'"),
        ):
            types.GridSettings.from_dict(data)

    def test_to_args(self):
        data = self.ok_values()
        assert types.GridSettings.from_dict(data).to_args() == {
            'search': 'foo',
            'onpage': 2,
            'perpage': 20,
            'op(test)': 'eq',
            'v1(test)': 'toast',
            'v2(test)': 'taft',
            'op(test2)': 'in',
            'v1(test2)': 'tarp',
            'sort1': 'bar',
            'sort2': '-baz',
            'export_to': None,
        }



from abc import ABC, abstractmethod
import decimal

from blazeutils import tolist

from .extensions import gettext as _


class ValueInvalid(Exception):
    def __init__(self, msg, value, instance):
        self.msg = msg
        self.value = value
        self.instance = instance

    def __str__(self):
        return self.msg


class Validator(ABC):
    @abstractmethod
    def process(self, value):
        pass


class StringValidator(Validator):
    def process(self, value):
        if value is None or value == '':
            return None
        if not isinstance(value, str):
            return str(value)
        return value


class RequiredValidator(Validator):
    def process(self, value):
        if value is None or value == '':
            raise ValueInvalid(_('Please enter a value.'), value, self)
        return value


class IntValidator(Validator):
    def process(self, value):
        if value is None or value == '':
            return None
        try:
            return int(value)
        except (ValueError, TypeError) as e:
            raise ValueInvalid(_('Please enter an integer value.'), value, self) from e


class FloatValidator(Validator):
    def process(self, value):
        if value is None or value == '':
            return None
        try:
            return float(value)
        except (ValueError, TypeError) as e:
            raise ValueInvalid(_('Please enter a number.'), value, self) from e


class DecimalValidator(Validator):
    def process(self, value):
        if value is None or value == '':
            return None
        try:
            return decimal.Decimal(value)
        except decimal.InvalidOperation as e:
            raise ValueInvalid(_('Please enter a number.'), value, self) from e


class RangeValidator(Validator):
    def __init__(self, min=None, max=None):  # noqa: A002
        if min is None and max is None:
            raise Exception(_('must specify either min or max for range validation'))
        self.min = min
        self.max = max

    def process(self, value):
        if self.min is not None and value is not None and value < self.min:
            raise ValueInvalid(
                _('Value must be greater than or equal to {}.').format(self.min),
                value,
                self,
            )
        if self.max is not None and value is not None and value > self.max:
            raise ValueInvalid(
                _('Value must be less than or equal to {}.').format(self.max),
                value,
                self,
            )
        return value


class OneOfValidator(Validator):
    def __init__(self, allowed_values):
        self.allowed_values = tuple(tolist(allowed_values))

    def process(self, value):
        if value is None or value == '':
            return None
        if value not in self.allowed_values:
            raise ValueInvalid(
                _('Value must be one of {}.').format(self.allowed_values),
                value,
                self,
            )
        return value


class CustomValidator(Validator):
    def __init__(self, processor=None):
        if not callable(processor):
            raise Exception(_('Processor should be callable and take a value argument'))
        self.processor = processor

    def process(self, value):
        return self.processor(value)



from os import environ
import sys


# Default URLs works for Docker compose and CI
db_kind = environ.get('WEBTEST_DB', 'pg')

if db_kind == 'pg':
    db_port = environ.get('DC_POSTGRES_PORT', '5432')
    default_url = f'postgresql+psycopg://postgres@127.0.0.1:{db_port}/postgres'
elif db_kind == 'mssql':
    # https://learn.microsoft.com/en-us/sql/connect/odbc/linux-mac/installing-the-microsoft-odbc-driver-for-sql-server
    db_port = environ.get('DC_MSSQL_PORT', '1433')
    default_url = f'mssql+pyodbc://SA:Docker-sa-password@127.0.0.1:{db_port}/tempdb?driver=ODBC+Driver+18+for+SQL+Server&trustservercertificate=yes'
else:
    assert db_kind == 'sqlite'
    default_url = 'sqlite:///'

db_url = environ.get('SQLALCHEMY_DATABASE_URI', default_url)

print('Webgrid sys.path', '\n'.join(sys.path))


def pytest_configure(config):
    from webgrid_ta.app import create_app

    app = create_app(config='Test', database_url=db_url)
    app.app_context().push()

    from webgrid_ta.model import load_db

    load_db()



import json

import flask

from webgrid import extensions, renderers


try:
    from morphi.helpers.jinja import configure_jinja_environment
except ImportError:
    configure_jinja_environment = lambda *args, **kwargs: None


class WebGrid(extensions.FrameworkManager):
    """Grid manager for connecting grids to Flask webapps.

    Manager is a Flask extension, and may be bound to an app via ``init_app``.

    Instance should be assigned to the manager attribute of a grid class::

        class MyGrid(BaseGrid):
            manager = WebGrid()

    Args:
        db (flask_sqlalchemy.SQLAlchemy, optional): Database instance. Defaults to None.
        If db is not supplied here, it can be set via ``init_db`` later.

    Class Attributes:
        jinja_loader (jinja.Loader): Template loader to use for HTML rendering.

        args_loaders (ArgsLoader[]): Iterable of classes to use for loading grid args, in order
        of priority

        session_max_hours (int): Hours to hold a given grid session in storage. Set to None to
        disable. Default 12.

        blueprint_name (string): Identifier to use for the Flask blueprint on this extension.
        Default "webgrid". Needs to be unique if multiple managers are initialized as flask
        extensions.
    """

    blueprint_name = 'webgrid'
    blueprint_class = flask.Blueprint

    def __init__(
        self,
        db=None,
        jinja_loader=None,
        args_loaders=None,
        session_max_hours=None,
        blueprint_name=None,
        blueprint_class=None,
    ):
        self.blueprint_name = blueprint_name or self.blueprint_name
        self.blueprint_class = blueprint_class or self.blueprint_class
        self._registered_grids = {}
        super().__init__(
            db=db,
            jinja_loader=jinja_loader,
            args_loaders=args_loaders,
            session_max_hours=session_max_hours,
        )

    def init_db(self, db):
        """Set the db connector."""
        self.db = db

    def sa_query(self, *args, **kwargs):
        """Wrap SQLAlchemy query instantiation."""
        return self.db.session.query(*args, **kwargs)

    def request_form_args(self):
        """Return POST request args."""
        return flask.request.form

    def request_json(self):
        """Return json body of request."""
        return flask.request.json

    def request_url_args(self):
        """Return GET request args."""
        return flask.request.args

    def csrf_token(self):
        """Return a CSRF token for POST."""
        from flask_wtf.csrf import generate_csrf

        return generate_csrf()

    def web_session(self):
        """Return current session."""
        return flask.session

    def persist_web_session(self):
        """Some frameworks require an additional step to persist session data."""
        flask.session.modified = True

    def flash_message(self, category, message):
        """Add a flash message through the framework."""
        flask.flash(message, category)

    def request(self):
        """Return request."""
        return flask.request

    def test_request_context(self, url='/'):
        """Get request context for tests."""
        return flask.current_app.test_request_context(url)

    def static_url(self, url_tail):
        """Construct static URL from webgrid blueprint."""
        return flask.url_for(f'{self.blueprint_name}.static', filename=url_tail)

    def init_blueprint(self, app):
        """Create a blueprint for webgrid assets."""
        return self.blueprint_class(
            self.blueprint_name,
            __name__,
            static_folder='static',
            static_url_path=app.static_url_path + '/webgrid',
        )

    def init_app(self, app):
        """Register a blueprint for webgrid assets, and configure jinja templates."""
        self.blueprint = self.init_blueprint(app)
        app.register_blueprint(self.blueprint)
        configure_jinja_environment(app.jinja_env, extensions.translation_manager)

    def file_as_response(self, data_stream, file_name, mime_type):
        """Return response from framework for sending a file."""
        as_attachment = file_name is not None
        return flask.send_file(
            data_stream,
            mimetype=mime_type,
            as_attachment=as_attachment,
            download_name=file_name,
        )


class WebGridAPI(WebGrid):
    """Subclass of WebGrid manager for creating an API connected to grid results.

    Manager is a Flask extension, and may be bound to an app via ``init_app``.

    Grids intended for API use should be registered on the manager via ``register_grid``.

    Security note: no attempt is made here to perform explicit authentication or
    authorization for the view. Those layers of functionality are the app developer's
    responsibility. For generic auth, ``api_view_method`` may be wrapped/overridden,
    or set up ``check_auth`` accordingly on your base grid class. Grid-specific auth can
    be handled in each grid's ``check_auth``.

    CSRF note: CSRF protection is standard security practice on Flask apps via
    ``flask_wtf``, but that is not really applicable to API endpoints. We take posts
    on the API view, but they will not inject information for storage - the API simply sets
    up and returns a grid. If you do wish to have CSRF protection, use the ``csrf_protection``
    class attribute and apply ``CSRFProtect`` to your application. Note that in that scenario
    you will need to apply the CSRF token manually to any API requests. If CSRF protection
    is not desired for the grid API endpoint, make sure ``init_app`` is called on the
    CSRF extension prior to this grid manager.

    Special Class Attributes:
        api_route_prefix (string): Prefix for URL route to bind on the manager's blueprint.
        Default "/webgrid-api". By default, ``api_route`` uses this to construct
        "/webgrid-api/<grid_ident>".
    """

    blueprint_name = 'webgrid-api'
    api_route_prefix = '/webgrid-api'
    args_loaders = (extensions.RequestJsonLoader,)
    csrf_protection = False

    def __init__(
        self,
        db=None,
        jinja_loader=None,
        args_loaders=None,
        session_max_hours=None,
        blueprint_name=None,
        blueprint_class=None,
        api_route_prefix=None,
    ):
        self.api_route_prefix = api_route_prefix or self.api_route_prefix
        self._registered_grids = {}
        super().__init__(
            db=db,
            jinja_loader=jinja_loader,
            args_loaders=args_loaders,
            session_max_hours=session_max_hours,
            blueprint_name=blueprint_name,
            blueprint_class=blueprint_class,
        )

    def init_blueprint(self, app):
        """Create a blueprint for webgrid assets and set up a generic API endpoint."""
        blueprint = super().init_blueprint(app)

        # conditionally exempt the blueprint from CSRF protection
        if not self.csrf_protection and app.extensions.get('csrf'):
            app.extensions['csrf'].exempt(blueprint)

        blueprint.route(self.api_route, methods=('POST',))(self.api_view_method)
        blueprint.route(self.api_route + '/count', methods=('POST',))(self.api_count_view_method)

        if app.config.get('TESTING'):

            @blueprint.route(self.api_route_prefix + '/testing/__csrf__', methods=('GET',))
            def csrf_get():
                from flask_wtf.csrf import generate_csrf

                return generate_csrf()

        return blueprint

    @property
    def api_route(self):
        """URL route to bind on the manager's blueprint for serving grids."""
        return self.api_route_prefix + '/<grid_ident>'

    def register_grid(self, grid_ident, grid_cls_or_creator):
        """Identify a grid class for API use via an identifying string.

        The identifier provided here will be used in route matching to init the
        requested grid. Identifiers are enforced as unique.

        ``grid_cls_or_creator`` may be a grid class or some other callable returning
        a grid instance.
        """
        if grid_ident in self._registered_grids:
            raise Exception('API grid_ident must be unique')

        self._registered_grids[grid_ident] = grid_cls_or_creator

    def api_init_grid(self, grid_cls_or_creator):
        """Create the grid instance from the registered class/creator."""
        return grid_cls_or_creator()

    def api_init_grid_post(self, grid):
        """Hook to run API-level init on every grid after instantiation"""

    def api_on_render_limit_exceeded(self, grid):
        """Export failed due to number of records. Returns a JSON response."""
        return flask.jsonify(error='too many records for render target')

    def api_export_response(self, grid):
        """Set up grid for export and return the response. Handles render limit exception."""
        import webgrid

        try:
            return grid.export_as_response()
        except webgrid.renderers.RenderLimitExceeded:
            return self.api_on_render_limit_exceeded(grid)

    def generate_requested_grid(self, grid_ident):
        if grid_ident not in self._registered_grids:
            flask.abort(404)

        grid = self.api_init_grid(self._registered_grids.get(grid_ident))
        self.api_init_grid_post(grid)

        # Make the API as flexible as possible to accept JSON post requests for grids
        # that may be used in other areas of the application, hence managed by a different
        # Flask extension/manager.
        grid.manager = self

        # Check auth before applying any args
        grid.check_auth()
        grid.apply_qs_args(add_user_warnings=False)

        return grid

    def api_view_method(self, grid_ident):
        """Main API view method. Returns JSON-rendered grid or desired export.

        No authentication/authorization is explicit here. Be sure to apply generic
        auth or set up ``check_auth`` on specific grids, if authorization is needed.

        If the ``grid_ident`` is not registered, response is 404.
        """
        grid = self.generate_requested_grid(grid_ident)

        if grid.export_to:
            return self.api_export_response(grid)

        # Be as flexible as possible here. If the grid has a JSON renderer, use it. But,
        # provide a default if it does not.
        renderer = getattr(grid, 'json', renderers.JSON(grid))

        # not using jsonify here because the JSON renderer returns a string
        return flask.Response(renderer(), mimetype='application/json')

    def api_count_view_method(self, grid_ident):
        """API view method to count records without returning the records and other grid info.
        Returns a dict with a count key.

        No authentication/authorization is explicit here. Be sure to apply generic
        auth or set up ``check_auth`` on specific grids, if authorization is needed.

        If the ``grid_ident`` is not registered, response is 404.
        """
        grid = self.generate_requested_grid(grid_ident)

        # not using jsonify here because the JSON renderer returns a string
        return flask.Response(json.dumps({'count': grid.record_count}), mimetype='application/json')



from abc import ABC, abstractmethod
import datetime
from decimal import Decimal
import json
from pathlib import Path
import re
from typing import Any
import warnings

import arrow
import jinja2 as jinja
from werkzeug.datastructures import MultiDict

from . import types


MORPHI_PACKAGE_NAME = 'webgrid'

# begin morphi boilerplate
try:
    import morphi
except ImportError:
    morphi = None

if morphi:
    from morphi.messages import Manager
    from morphi.registry import default_registry

    translation_manager = Manager(package_name=MORPHI_PACKAGE_NAME)
    default_registry.subscribe(translation_manager)

    gettext = translation_manager.gettext
    lazy_gettext = translation_manager.lazy_gettext
    lazy_ngettext = translation_manager.lazy_ngettext
    ngettext = translation_manager.ngettext

else:
    translation_manager = None

    def gettext(message, **variables):
        if variables:
            return message.format(**variables)

        return message

    def ngettext(singular, plural, num, **variables):
        variables.setdefault('num', num)

        if num == 1:
            return gettext(singular, **variables)

        return gettext(plural, **variables)

    lazy_gettext = gettext
    lazy_ngettext = ngettext


class CustomJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime.date, arrow.Arrow)):
            return obj.isoformat()
        elif isinstance(obj, Decimal):
            return float(obj)

        try:
            return super().default(obj)
        except TypeError:
            return str(obj)


class ArgsLoader(ABC):
    """Base args loader class.

    When a grid calls for its args, it requests them from the manager. The manager will have one
    or more args loaders to be run in order. Each loader fetches its args from the request,
    then ensuing loaders have the opportunity to modify or perform other operations as needed.
    """

    def __init__(self, manager):
        self.manager = manager

    @abstractmethod
    def get_args(self, grid, previous_args):
        pass


class GridPrefixBase(ArgsLoader):
    def sanitize_arg_name(self, arg_name, qs_prefix):
        if qs_prefix and arg_name.startswith(qs_prefix):
            return arg_name[len(qs_prefix) :]
        return arg_name

    def get_sanitized_args(self, grid, args):
        incoming_args = []
        if grid.qs_prefix:
            for key, values in MultiDict(args).lists():
                # Only include args that start with the grid's prefix
                if not key.startswith(grid.qs_prefix):
                    continue
                key = self.sanitize_arg_name(key, grid.qs_prefix)
                for single_val in values:
                    incoming_args.append((key, single_val))
        else:
            incoming_args = args

        return MultiDict(incoming_args)

    def get_args(self, grid, previous_args):
        request_args = self.get_args_from_request()
        if 'dgreset' in request_args:
            if 'session_key' in request_args:
                return MultiDict({'dgreset': 1, 'session_key': request_args['session_key']})
            return MultiDict({'dgreset': 1})

        request_args = self.get_sanitized_args(grid, request_args)
        request_args.update(previous_args)
        return request_args

    @abstractmethod
    def get_args_from_request(self):
        pass


class RequestArgsLoader(GridPrefixBase, ArgsLoader):
    """Simple args loader for web request.

    Args are usually passed through directly from the request. If the grid has a query string
    prefix, the relevant args will be namespaced - sanitize them here and return the subset
    needed for the given grid.

    In the reset case, ignore most args, and return only the reset flag and session key (if any).
    """

    def get_args_from_request(self):
        return self.manager.request_url_args()


class RequestFormLoader(GridPrefixBase, ArgsLoader):
    """Simple form loader for web request.

    Form values are usually passed through directly from the request. If the grid has a prefix, the
    relevant args will be namespaced - sanitize them here and return the subset needed for the given
    grid.

    In the reset case, ignore most args, and return only the reset flag and session key (if any).
    """

    def get_args_from_request(self):
        return self.manager.request_form_args()


class RequestJsonLoader(ArgsLoader):
    """JSON loader for web request.

    See :meth:`webgrid.types.GridSettings` for the expected JSON structure. The parsed arguments
    are converted to the querystring arg format and merged with any previous args.
    """

    def json_to_args(self, data: dict[str, Any]):
        meta = types.GridSettings.from_dict(data)
        return MultiDict(meta.to_args())

    def get_args(self, grid, previous_args):
        data = self.manager.request_json()
        current_args = self.json_to_args(data) if data else MultiDict()
        current_args.update(previous_args)
        return current_args


class WebSessionArgsLoader(ArgsLoader):
    """Session manager for grid args.

    Args are assumed to have been sanitized from the request already. But, args may be combined
    from the request and the session store for a number of cases.

    - If session_override or no filter ops, load args from session store
    - Having filter ops present will reset session store unless session_override is present
    - Page/sort args will always take precedence over stored args, but not reset the store
    - Export argument is handled outside of the session store

    In the reset case, ignore most args, and return only the reset flag and session key (if any).
    And clear the session store for the given grid.
    """

    _session_exclude_keys = (
        '__foreign_session_loaded__',
        'apply',
        'dgreset',
        'export_to',
        'session_key',
        'session_override',
    )

    def args_have_op(self, args):
        """Check args for containing any filter operators.

        Args:
            args (MultiDict): Request args.

        Returns:
            bool: True if at least one op is present.
        """
        regex = re.compile(r'(op\(.*\))')
        return any(regex.match(a) for a in args)

    def args_have_page(self, args):
        """Check args for containing any page args.

        Args:
            args (MultiDict): Request args.

        Returns:
            bool: True if at least one page arg is present.
        """
        regex = re.compile('(onpage|perpage)')
        return any(regex.match(a) for a in args)

    def args_have_sort(self, args):
        """Check args for containing any sort keys.

        Args:
            args (MultiDict): Request args.

        Returns:
            List[str]: all args matching as sort args.
        """
        regex = re.compile('(sort[1-9][0-9]*)')
        return [
            arg.string
            for arg in filter(
                lambda match: match is not None,
                [regex.match(a) for a in args],
            )
        ]

    def remove_grid_session(self, session_key):
        # Remove a grid session from the cookie entirely
        if 'dgsessions' not in self.manager.web_session():
            return
        if session_key not in self.manager.web_session()['dgsessions']:
            return

        self.manager.web_session()['dgsessions'].pop(session_key)
        self.manager.persist_web_session()

    def apply_session_overrides(self, session_args, previous_args):
        """Update session args as needed from the incoming request.

        If session override case, wholesale update from the incoming request. This
        is useful if a single filter needs to be changed via the URL, but we don't
        want to dump the rest of the stored filters from the session.

        Otherwise, apply only page/sort if available in the request.

        Export directive is passed through from the request, so a session store never
        triggers an export by itself.

        Args:
            session_args (MultiDict): Args loaded from the session store.
            previous_args (MultiDict): Args that came into this args loader.

        Returns:
            MultiDict: Args to be used in grid operations.
        """
        is_override = 'session_override' in previous_args

        if is_override:
            # Typically we would want to just `.update()`, but args are MultiDict, so
            # could have multiple values already. Previous args values are to be
            # substituted, not treated as a list extension.
            # Using copy, because if session_args and previous_args are the same
            # object, we would be modifying the MultiDict in-place.
            for key, value in previous_args.copy().lists():
                session_args.setlist(key, value)
        else:
            # Some types of args always get passed through from request_args.
            # Override paging if it exists in the query
            if self.args_have_page(previous_args):
                session_args['onpage'] = previous_args.get('onpage')
                session_args['perpage'] = previous_args.get('perpage')
            # Override sorting if it exists in the query
            for sort_arg in self.args_have_sort(previous_args):
                session_args[sort_arg] = previous_args.get(sort_arg)

        if 'export_to' in previous_args:
            # Export directive gets left out of the session storage, since a request may
            # have session key and export, and filter/sort must be loaded.
            session_args['export_to'] = previous_args['export_to']

        return session_args

    def cleanup_expired_sessions(self):
        """Remove sessions older than a certain number of hours.

        Configurable at the manager level, with the session_max_hours attribute. If
        None, cleanup is disabled.
        """
        if self.manager.session_max_hours is None:
            return

        cutoff = arrow.utcnow().shift(hours=-self.manager.session_max_hours)
        modified = False

        web_session = self.manager.web_session()
        for session_key in tuple(web_session.get('dgsessions', {}).keys()):
            grid_session = self.get_session_store(None, {'session_key': session_key})
            if arrow.get(grid_session.get('session_timestamp', cutoff)) < cutoff:
                modified = True
                web_session['dgsessions'].pop(session_key, None)

        if modified:
            self.manager.persist_web_session()

    def get_session_store(self, grid, args):
        """Load args from session by session_key, and return as MultiDict.

        Args:
            grid (BaseGrid): Grid used to get default grid key (based on grid class name).
            args (MultiDict): Request args used for session key.

        Returns:
            MultiDict: Args to be used in grid operations.
        """
        # check args for a session key. If the key is present,
        #   look it up in the session and use the saved args
        #   (if they have been saved under that key). If not,
        #   look up the class name for a default arg store.
        web_session = self.manager.web_session()
        if 'dgsessions' not in web_session:
            return args
        dgsessions = web_session['dgsessions']

        # session is stored as a JSON-serialized list of tuples, which we can turn into MultiDict
        session_key = grid.default_session_key if grid else None
        grid_session_key = args.get('session_key', None)
        if grid_session_key and dgsessions.get(grid_session_key):
            session_key = grid_session_key
        stored_args_json = dgsessions.get(session_key, '[]')
        if isinstance(stored_args_json, MultiDict):
            stored_args_json = json.dumps(list(stored_args_json.items(multi=True)))
        elif isinstance(stored_args_json, dict):
            stored_args_json = json.dumps(list(stored_args_json.items()))
        return MultiDict(json.loads(stored_args_json))

    def save_session_store(self, grid, args):
        """Save the args in the session under the session key and as defaults for this grid.

        Note, export and reset args are ignored for storage.

        Args:
            args (MultiDict): Request args to be loaded in next session store retrieval.
        """
        # save the args in the session under the session key
        #   and also as the default args for this grid
        web_session = self.manager.web_session()
        if 'dgsessions' not in web_session:
            web_session['dgsessions'] = {}
        dgsessions = web_session['dgsessions']
        grid_session_key = args.get('session_key') or grid.session_key
        # work with a copy here
        args = MultiDict(args)
        # remove keys that should not be stored
        for key in self._session_exclude_keys:
            args.pop(key, None)

        existing_default_store = self.get_session_store(grid, MultiDict([]))

        # if we're only storing the bare minimal case, remove the store, including the default
        if not args:
            dgsessions.pop(grid_session_key, None)
            dgsessions.pop(grid.default_session_key, None)
            return self.manager.persist_web_session()

        args['datagrid'] = grid.default_session_key
        args['session_timestamp'] = existing_default_store['session_timestamp'] = (
            arrow.utcnow().isoformat()
        )

        # if we're pulling a grid matching the default session, but with a different key,
        # no need to store the sepearate session
        if args == existing_default_store:
            dgsessions.pop(grid_session_key, None)
            return self.manager.persist_web_session()

        # serialize the args so we can enforce the correct MultiDict type on the other side
        args_json = json.dumps(list(args.items(multi=True)))
        # save in store under grid default and session key
        dgsessions[grid_session_key] = args_json
        dgsessions[grid.default_session_key] = args_json

        # some frameworks/sessions need these changes manually persisted
        self.manager.persist_web_session()

    def get_args(self, grid, previous_args):
        """Retrieve args from session and override as appropriate.

        Submitting the header form flushes all args to the URL, so no need to load them
        from session.

        If that is not the path, then we either have filtering args on the URL, or not. Default
        behavior is currently to consider filtering args to be comprehensive and authoritative,
        UNLESS a session_override arg is present.

        The special session_override arg causes the store to overlay request args against an
        existing session, and return the combination.

        Args:
            grid (BaseGrid): Grid used to get default grid key (based on grid class name).
            previous_args (MultiDict): Incoming args, assumed to be sanitized already.

        Returns:
            MultiDict: Args to be used in grid operations.
        """
        self.cleanup_expired_sessions()

        if not grid.session_on:
            # Shouldn't be a normal case anymore. But if the grid has session store disabled,
            # honor that and just return the args that came in.
            return previous_args

        if 'dgreset' in previous_args:
            # Request has a reset. Do nothing else.
            if grid.session_on:
                self.remove_grid_session(previous_args.get('session_key') or grid.session_key)
                self.remove_grid_session(grid.default_session_key)
            return MultiDict({'dgreset': 1, 'session_key': previous_args.get('session_key')})

        # From here on, work with a copy so as not to mutate the incoming args
        request_args = previous_args.copy()

        is_override = 'session_override' in request_args
        is_apply = 'apply' in request_args
        if (not self.args_have_op(request_args) and not is_apply) or is_override:
            # We are in a session-loading case.
            session_args = self.apply_session_overrides(
                self.get_session_store(grid, request_args),
                request_args,
            )

            # Flag a foreign session if loading from another grid's session.
            session_args['__foreign_session_loaded__'] = (
                session_args.get('datagrid', grid.default_session_key) != grid.default_session_key
            )

            # No further need to treat args lists separately.
            request_args = session_args

        self.save_session_store(grid, request_args)
        return request_args


class FrameworkManager(ABC):
    """Grid manager base class for connecting grids to webapps.

    Provides common framework-related properties and methods.

    Class Attributes:
        jinja_loader (jinja.Loader): Template loader to use for HTML rendering, or a lambda
        function to create one

        args_loaders (ArgsLoader[]): Iterable of classes to use for loading grid args, in order
        of priority

        session_max_hours (int): Hours to hold a given grid session in storage. Set to None to
        disable. Default 12.

    """

    jinja_loader = lambda self: jinja.PackageLoader('webgrid', 'templates')
    args_loaders = (
        RequestArgsLoader,
        WebSessionArgsLoader,
    )
    session_max_hours = 12

    def __init__(self, db=None, jinja_loader=None, args_loaders=None, session_max_hours=None):
        self.init_db(db)

        self.jinja_loader = jinja_loader or self.jinja_loader
        if not isinstance(self.jinja_loader, jinja.BaseLoader) and callable(self.jinja_loader):
            self.jinja_loader = self.jinja_loader()
        self.args_loaders = args_loaders or self.args_loaders
        if session_max_hours is not None:
            # condition must account for possibility of 0 being passed in
            self.session_max_hours = session_max_hours

        self.init_jinja()

        if callable(getattr(self, 'init', False)):
            self.init()

    def init_db(self, db):
        """Set the db connector."""
        self.db = db

    def init_jinja(self):
        """Configure grid-specific jinja environment."""
        self.jinja_environment = jinja.Environment(
            loader=self.jinja_loader,
            finalize=lambda x: x if x is not None else '',
            autoescape=True,
        )

    def static_path(self):
        """Path where static files are located on the filesystem."""
        return str(Path(__file__).resolve().parent / 'static')

    def get_args(self, grid):
        args = MultiDict()
        """Run request args through manager's args loaders, and return the result."""
        for loader in self.args_loaders:
            args = loader(self).get_args(grid, args)

        return args

    @abstractmethod
    def test_request_context(self, url='/'):
        """Get request context for tests."""

    def request_args(self):
        warnings.warn(
            'request_args is deprecated and will be removed in a future version.',
            DeprecationWarning,
            2,
        )
        return self.request_url_args()

    @abstractmethod
    def request_json(self):
        """Return json body of request."""

    @abstractmethod
    def request_form_args(self):
        """Return GET request args."""

    @abstractmethod
    def request_url_args(self):
        """Return POST request args."""

    @abstractmethod
    def csrf_token(self):
        """Return a CSRF token for POST."""



from blazeweb.views import View

from webgrid import NumericColumn
from webgrid_blazeweb_ta.extensions import lazy_gettext as _
from webgrid_blazeweb_ta.model.orm import Person
from webgrid_blazeweb_ta.tests.grids import PeopleGrid as PGBase


class CurrencyCol(NumericColumn):
    def format_data(self, data):
        return data if int(data) % 2 else data * -1


class PeopleGrid(PGBase):
    CurrencyCol(_('Currency'), Person.numericcol, format_as='percent', places=5)
    CurrencyCol(_('C2'), Person.numericcol.label('n2'), format_as='accounting')


class ManagePeople(View):
    def default(self):
        pg = PeopleGrid()
        pg.apply_qs_args()
        if pg.export_to:
            pg.export_as_response()
        self.assign('people_grid', pg)
        self.render_template()



from pathlib import Path

from blazeutils.testing import assert_equal_txt
import flask
from flask_webtest import SessionScope
import sqlalchemy as sa
from werkzeug.datastructures import MultiDict
import wrapt

from webgrid import Column
from webgrid_ta.model import db


cdir = Path(__file__).parent

db_sess_scope = SessionScope(db)


class ModelBase:
    @classmethod
    def setup_class(cls):
        db_sess_scope.push()

    @classmethod
    def teardown_class(cls):
        db_sess_scope.pop()


def eq_html(html, filename):
    with cdir.joinpath('data', filename).open('rb') as fh:
        file_html = fh.read().decode('ascii')
    assert_equal_txt(html, file_html)


def _inrequest(*req_args, **req_kwargs):
    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        with flask.current_app.test_request_context(*req_args, **req_kwargs):
            # replaces request.args wth MultiDict so it is mutable
            flask.request.args = MultiDict(flask.request.args)
            return wrapped(*args, **kwargs)

    return wrapper


def render_in_grid(grid_cls, render_in):
    """Class factory which extends an existing grid class
    to add a column that is rendered everywhere except "render_in"
    """
    other_render_types = set(Column._render_in)
    other_render_types.remove(render_in)

    class RenderInGrid(grid_cls):
        Column('Exclude', sa.literal('Exclude'), render_in=tuple(other_render_types))

    return RenderInGrid



from pathlib import Path

from blazeutils.testing import assert_equal_txt
import sqlalchemy.orm


cdir = Path(__file__).parent


def query_to_str(statement, bind=None):
    """
    returns a string of a sqlalchemy.orm.Query with parameters bound

    WARNING: this is dangerous and ONLY for testing, executing the results
    of this function can result in an SQL Injection attack.
    """
    if isinstance(statement, sqlalchemy.orm.Query):
        if bind is None:
            bind = statement.session.get_bind()
        statement = statement.statement
    elif bind is None:
        bind = statement.bind

    if bind is None:
        raise Exception(
            'bind param (engine or connection object) required when using with an unbound statement',  # noqa: E501
        )

    dialect = bind.dialect
    compiler = statement._compiler(dialect)

    class LiteralCompiler(compiler.__class__):
        def visit_bindparam(
            self,
            bindparam,
            within_columns_clause=False,
            literal_binds=False,
            **kwargs,
        ):
            return super().render_literal_bindparam(
                bindparam,
                within_columns_clause=within_columns_clause,
                literal_binds=literal_binds,
                **kwargs,
            )

    compiler = LiteralCompiler(dialect, statement)
    return 'TESTING ONLY BIND: ' + compiler.process(statement)


def eq_html(html, filename):
    with cdir.joinpath('data', filename).open('rb') as fh:
        file_html = fh.read()
    assert_equal_txt(html, file_html)


def assert_in_query(obj, test_for):
    query = obj.build_query() if hasattr(obj, 'build_query') else obj
    query_str = query_to_str(query)
    assert test_for in query_str, query_str


def assert_not_in_query(obj, test_for):
    query = obj.build_query() if hasattr(obj, 'build_query') else obj
    query_str = query_to_str(query)
    assert test_for not in query_str, query_str



from dataclasses import dataclass, field
from typing import Any


class ValidationError(Exception):
    pass


class FieldValidationError(ValidationError):
    def __init__(self, field, value, type_):
        message = f'Received {field}={value}; should be of type {type_}'
        super().__init__(message)


@dataclass
class Filter:
    op: str
    value1: str | list[str]
    value2: str | None = None


@dataclass
class Paging:
    pager_on: bool = False
    per_page: int | None = None
    on_page: int | None = None

    def __post_init__(self):
        if self.per_page is not None and not isinstance(self.per_page, int):
            raise FieldValidationError('per_page', self.per_page, 'int')
        if self.on_page is not None and not isinstance(self.on_page, int):
            raise FieldValidationError('on_page', self.on_page, 'int')


@dataclass
class Sort:
    key: str
    flag_desc: bool


@dataclass
class FilterOperator:
    key: str
    label: str
    field_type: str | None
    hint: str | None = None


@dataclass
class FilterOption:
    key: str
    value: str


@dataclass
class FilterSpec:
    operators: list[FilterOperator]
    primary_op: FilterOperator | None


@dataclass
class OptionsFilterSpec(FilterSpec):
    options: list[FilterOption]


@dataclass
class ColumnGroup:
    label: str
    columns: list[str]


@dataclass
class GridTotals:
    page: dict[str, Any] | None = None
    grand: dict[str, Any] | None = None


@dataclass
class GridSettings:
    search_expr: str | None = None
    filters: dict[str, Filter] = field(default_factory=dict)
    paging: Paging = field(default_factory=Paging)
    sort: list[Sort] = field(default_factory=list)
    export_to: str | None = None

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> 'GridSettings':
        """Create from deserialized json"""
        try:
            filters = {key: Filter(**filter_) for key, filter_ in data.get('filters', {}).items()}
        except TypeError as e:
            raise ValidationError(f'Filter: {e}') from e

        try:
            paging = Paging(**data.get('paging', {}))
        except TypeError as e:
            raise ValidationError(f'Paging: {e}') from e

        try:
            sort = [Sort(**sort) for sort in data.get('sort', [])]
        except TypeError as e:
            raise ValidationError(f'Sort: {e}') from e

        return cls(
            search_expr=data.get('search_expr'),
            filters=filters,
            paging=paging,
            sort=sort,
            export_to=data.get('export_to'),
        )

    def to_args(self) -> dict[str, Any]:
        """Convert grid parameters to request args format"""
        args = {
            'search': self.search_expr,
            'onpage': self.paging.on_page,
            'perpage': self.paging.per_page,
            'export_to': self.export_to,
        }

        for key, filter_ in self.filters.items():
            args[f'op({key})'] = filter_.op
            args[f'v1({key})'] = filter_.value1
            if filter_.value2:
                args[f'v2({key})'] = filter_.value2

        for i, s in enumerate(self.sort, 1):
            prefix = '-' if s.flag_desc else ''
            args[f'sort{i}'] = f'{prefix}{s.key}'

        return args


@dataclass
class GridSpec:
    columns: list[dict[str, str]]
    column_groups: list[ColumnGroup]
    column_types: list[dict[str, str]]
    export_targets: list[str]
    enable_search: bool
    enable_sort: bool
    sortable_columns: list[str]
    filters: dict[str, FilterSpec] = field(default_factory=dict)


@dataclass
class GridState:
    page_count: int
    record_count: int
    warnings: list[str]


@dataclass
class Grid:
    settings: GridSettings
    spec: GridSpec
    state: GridState
    records: list[dict[str, Any]]
    totals: GridTotals
    errors: list[str]



import datetime as dt
from decimal import Decimal
import json
from os import path
import re
from typing import ClassVar
from unittest import mock
from unittest.mock import MagicMock

import arrow
import flask
import pytest
import sqlalchemy.sql as sasql
from werkzeug.datastructures import MultiDict

from webgrid import BoolColumn, Column, YesNoColumn
from webgrid.extensions import (
    CustomJsonEncoder,
    RequestArgsLoader,
    RequestFormLoader,
    RequestJsonLoader,
    WebSessionArgsLoader,
)
from webgrid.extensions import (
    lazy_gettext as _,
)
from webgrid.filters import AggregateIntFilter, FilterBase, IntFilter, TextFilter
from webgrid.renderers import CSV
from webgrid.testing import assert_in_query, assert_not_in_query
from webgrid_ta.grids import Grid, PeopleGrid, PeopleGridByConfig
from webgrid_ta.model.entities import Person, Status, Stopwatch, db

from .helpers import _inrequest


class TestGrid:
    class TG(Grid):
        Column('First Name', Person.firstname)

    class KeyGrid(Grid):
        # demonstrate grid operations when multiple column expressions have the same key
        Column('Person ID', Person.id, IntFilter)
        Column('Stopwatch ID', Stopwatch.id, IntFilter)
        Column('No Expression')
        Column(_('No Expression'))
        Column('')

        query_joins: ClassVar = [(Stopwatch, Stopwatch.id > 0)]

    def setup_method(self, _):
        Status.delete_cascaded()
        Person.testing_create()
        Person.testing_create()
        Person.testing_create()
        Person.testing_create()
        Person.testing_create()

    def test_static_path(self):
        g = self.TG()
        assert g.manager.static_path().endswith(f'webgrid{path.sep}static')

    def test_column_key(self):
        g = self.TG()
        g.build_query()
        assert g.columns[0].key == 'firstname'

    def test_columns_are_new_instances(self):
        g = self.TG()
        g2 = self.TG()
        assert g.columns[0] is not self.TG.__cls_cols__[0]
        assert g.columns[0].__class__ is self.TG.__cls_cols__[0].__class__
        assert g.columns[0] is not g2.columns[0]
        assert g.columns[0] is g.key_column_map['firstname']

    def test_grid_ident(self):
        class Grid1(Grid):
            identifier = 'cars'

        class Grid2(Grid):
            pass

        mg = Grid2()
        assert mg.ident == 'grid2'

        mg = Grid1()
        assert mg.ident == 'cars'

        mg = Grid1('foo')
        assert mg.ident == 'foo'

    def test_base_query(self):
        class CTG(Grid):
            Column('First Name', Person.firstname)

        g = CTG()
        query = g.build_query()
        assert_not_in_query(query, 'WHERE')
        if db.engine.dialect.name != 'mssql':
            assert_not_in_query(query, 'ORDER BY')
        else:
            # MSSQL queries get an ORDER BY patched in if none is provided,
            # else paging doesn't work
            assert_in_query(query, 'ORDER BY')
        with mock.patch('logging.Logger.debug') as m_debug:
            rs = g.records
            assert len(rs) > 0, rs
            expected = [
                r'^<Grid "CTG">$',
                r'^No filters$',
                r'^No sorts$',
                r'^Page 1; 50 per page$',
                r'^Data query ran in \d+\.?\d* seconds$',
            ]
            assert len(expected) == len(m_debug.call_args_list)
            for idx, call in enumerate(m_debug.call_args_list):
                assert re.match(expected[idx], call[0][0])

    def test_subtotal_computed(self):
        def TestGrid(subtotals_val):
            class CTG(Grid):
                subtotals = subtotals_val
                Column('Sum Total', Person.numericcol.label('something'), has_subtotal='sum')

            return CTG

        Person.testing_create(numericcol=5)
        Person.testing_create(numericcol=10)

        assert TestGrid('none')().grand_totals is None
        assert TestGrid('none')().page_totals is None
        assert TestGrid('grand')().page_totals is None
        assert TestGrid('page')().grand_totals is None
        assert TestGrid('grand')().grand_totals.something == 15
        assert TestGrid('page')().page_totals.something == 15
        assert TestGrid('all')().grand_totals.something == 15
        assert TestGrid('all')().page_totals.something == 15

    @mock.patch('logging.Logger.debug')
    def test_subtotal_sum_by_default(self, m_debug):
        class CTG(Grid):
            subtotals = 'grand'
            Column('Sum Total', Person.numericcol.label('something'), has_subtotal=True)

        Person.testing_create(numericcol=5)
        Person.testing_create(numericcol=10)
        g = CTG()
        totals = g.grand_totals
        assert totals.something == 15
        expected = [
            r'^<Grid "CTG">$',
            r'^No filters$',
            r'^Totals query ran in \d+\.?\d* seconds$',
        ]
        assert len(expected) == len(m_debug.call_args_list)
        for idx, call in enumerate(m_debug.call_args_list):
            assert re.match(expected[idx], call[0][0])

    def test_subtotal_sum(self):
        class CTG(Grid):
            subtotals = 'grand'
            Column('Sum Total', Person.numericcol.label('something'), has_subtotal='sum')

        Person.testing_create(numericcol=5)
        Person.testing_create(numericcol=10)
        g = CTG()
        totals = g.grand_totals
        assert totals.something == 15

    def test_subtotal_avg(self):
        class CTG(Grid):
            subtotals = 'grand'
            Column('Sum Total', Person.numericcol.label('something'), has_subtotal='avg')

        Person.testing_create(numericcol=5)
        Person.testing_create(numericcol=10)
        g = CTG()
        totals = g.grand_totals
        assert totals.something == Decimal('7.5')

    def test_subtotal_expr_string(self):
        class CTG(Grid):
            subtotals = 'grand'
            ratio_expr = Person.numericcol / Person.sortorder
            Column('Numeric', Person.numericcol.label('numeric_col'), has_subtotal=True)
            Column('Ints', Person.floatcol.label('float_col'), has_subtotal=True)
            Column(
                'Ratio',
                Person.numericcol.label('something'),
                has_subtotal='sum(numeric_col) / sum(float_col)',
            )

        Person.testing_create(numericcol=5, floatcol=1)
        Person.testing_create(numericcol=10, floatcol=3)
        g = CTG()
        totals = g.grand_totals
        assert totals.something == Decimal('3.75'), totals

    def test_subtotal_expr(self):
        sum_ = sasql.functions.sum

        class CTG(Grid):
            subtotals = 'grand'
            ratio_expr = Person.numericcol / Person.sortorder
            Column('Numeric', Person.numericcol.label('numeric_col'), has_subtotal=True)
            Column('Ints', Person.floatcol.label('float_col'), has_subtotal=True)
            Column(
                'Ratio',
                Person.numericcol.label('something'),
                has_subtotal=sum_(Person.numericcol) / sum_(Person.floatcol),
            )

        Person.testing_create(numericcol=5, floatcol=1)
        Person.testing_create(numericcol=10, floatcol=3)
        g = CTG()
        totals = g.grand_totals
        assert totals.something == Decimal('3.75'), totals

    def test_query_prep_sorting(self):
        class CTG(Grid):
            Column('First Name', Person.firstname)

            def query_prep(self, query, has_sort, has_filters):
                assert not has_sort
                assert not has_filters
                return query.order_by(Person.lastname)

        g = CTG()
        assert_in_query(g, 'ORDER BY persons.last_name')

    def test_column_reorder(self):
        class CTG(Grid):
            Column('First Name', Person.firstname, TextFilter)
            Column('Last Name', Person.lastname, TextFilter)

        grid = CTG()
        grid.set_column_order(('lastname', 'firstname'))
        assert grid.columns[-1].label == 'First Name'
        assert grid.columns[0].label == 'Last Name'

        with pytest.raises(Exception, match="Keys not recognized on grid: {'foo'}"):
            grid.set_column_order(('foo',))

    def test_add_column(self):
        class CTG(Grid):
            Column('First Name', Person.firstname, TextFilter)

        grid = CTG()
        grid.add_column(Column('Last Name', Person.lastname, TextFilter))

        assert len(grid.columns) == 2
        colinst = grid.column('lastname')
        assert colinst.grid is grid
        assert colinst.key == 'lastname'

    def test_filter_class(self):
        class CTG(Grid):
            Column('First Name', Person.firstname, TextFilter)
            Column('Last Name', Person.lastname, TextFilter)

        g = CTG()
        g.set_filter('firstname', 'eq', 'foo')
        if db.engine.dialect.name == 'mssql':
            assert_in_query(g, "WHERE persons.firstname = 'foo'")
        else:
            assert_in_query(g, "WHERE upper(persons.firstname) = upper('foo')")

        with mock.patch('logging.Logger.debug') as m_debug:
            g.records  # noqa: B018
            g.set_filter('lastname', 'eq', 'bar')
            g.records  # noqa: B018
            expected = [
                r'^<Grid "CTG">$',
                r'^firstname: class=TextFilter, op=eq, value1=foo, value2=None$',
                r'^No sorts$',
                r'^Page 1; 50 per page$',
                r'^Data query ran in \d+\.?\d* seconds$',
                r'^<Grid "CTG">$',
                r'^firstname: class=TextFilter, op=eq, value1=foo, value2=None;'
                r'lastname: class=TextFilter, op=eq, value1=bar, value2=None$',
                r'^No sorts$',
                r'^Page 1; 50 per page$',
                r'^Data query ran in \d+\.?\d* seconds$',
            ]
            assert len(expected) == len(m_debug.call_args_list)
            for idx, call in enumerate(m_debug.call_args_list):
                assert re.match(expected[idx], call[0][0])

    def test_filter_instance(self):
        class CTG(Grid):
            Column('First Name', Person.firstname, TextFilter(Person.lastname))

        g = CTG()
        g.set_filter('firstname', 'eq', 'foo')
        if db.engine.dialect.name == 'mssql':
            assert_in_query(g, "WHERE persons.last_name = 'foo'")
        else:
            assert_in_query(g, "WHERE upper(persons.last_name) = upper('foo')")

    def test_filter_two_values(self):
        class CTG(Grid):
            Column('Sort Order', Person.sortorder, IntFilter(Person.sortorder))

        g = CTG()
        g.set_filter('sortorder', 'between', 5, value2=10)
        assert_in_query(g, 'WHERE persons.sortorder BETWEEN 5 AND 10')

    def test_order_by(self):
        class CTG(Grid):
            Column('First Name', Person.firstname)

        g = CTG()
        g.set_sort('-firstname')
        assert_in_query(g, 'ORDER BY persons.firstname')
        with mock.patch('logging.Logger.debug') as m_debug:
            g.records  # noqa: B018
            expected = [
                r'^<Grid "CTG">$',
                r'^No filters$',
                r'^firstname$',
                r'^Page 1; 50 per page$',
                r'^Data query ran in \d+\.?\d* seconds$',
            ]
            assert len(expected) == len(m_debug.call_args_list)
            for idx, call in enumerate(m_debug.call_args_list):
                re.match(expected[idx], call[0][0])

    def test_redundant_order_by(self):
        class CTG(Grid):
            Column('First Name', Person.firstname)
            Column('Last Name', Person.lastname)

        g = CTG()
        g.set_sort('firstname', 'lastname', '-firstname')
        assert_in_query(g, 'ORDER BY persons.firstname, persons.last_name')
        assert_not_in_query(g, 'ORDER BY persons.firstname, persons.last_name,')
        with mock.patch('logging.Logger.debug') as m_debug:
            g.records  # noqa: B018
            expected = [
                r'^<Grid "CTG">$',
                r'^No filters$',
                r'^firstname,lastname$',
                r'^Page 1; 50 per page$',
                r'^Data query ran in \d+\.?\d* seconds$',
            ]
            assert len(expected) == len(m_debug.call_args_list)
            for idx, call in enumerate(m_debug.call_args_list):
                assert re.match(expected[idx], call[0][0])

    def test_paging(self):
        g = self.TG()
        if db.engine.dialect.name == 'mssql':
            assert_in_query(g, 'OFFSET 0 ROWS')
            assert_in_query(g, 'FETCH FIRST 50 ROWS ONLY')
        else:
            assert_in_query(g, 'LIMIT 50 OFFSET 0')

        g.set_paging(2, 1)
        if db.engine.dialect.name == 'mssql':
            assert_in_query(g, 'OFFSET 0 ROWS')
            assert_in_query(g, 'FETCH FIRST 2 ROWS ONLY')
        else:
            assert_in_query(g, 'LIMIT 2 OFFSET 0')

        g.set_paging(10, 5)
        if db.engine.dialect.name == 'mssql':
            assert_in_query(g, 'OFFSET 40 ROWS')
            assert_in_query(g, 'FETCH FIRST 10 ROWS ONLY')
        else:
            assert_in_query(g, 'LIMIT 10 OFFSET 40')

    def test_paging_disabled(self):
        class TG(Grid):
            pager_on = False
            Column('First Name', Person.firstname)

        g = TG()
        assert_not_in_query(g, 'LIMIT 50 OFFSET 0')
        assert g.page_count == 1

    def test_page_count(self):
        g = self.TG()
        assert g.page_count == 1

        g.set_paging(1, 2)
        assert g.page_count == 5

        g = self.TG(per_page=None)
        assert g.page_count == 1

    @mock.patch('logging.Logger.debug')
    def test_record_count(self, m_debug):
        g = self.TG()
        assert g.record_count == 5
        expected = [
            r'^<Grid "TG">$',
            r'^No filters$',
            r'^Count query ran in \d+\.?\d* seconds$',
        ]
        assert len(expected) == len(m_debug.call_args_list)
        for idx, call in enumerate(m_debug.call_args_list):
            assert re.match(expected[idx], call[0][0])

    @mock.patch('logging.Logger.debug')
    def test_record_count_preserved_during_sort(self, m_debug):
        g = self.TG()
        assert g.record_count == 5
        assert m_debug.call_count == 3
        g.set_sort('firstname')
        assert g.record_count == 5
        assert m_debug.call_count == 3

    @mock.patch('logging.Logger.debug')
    def test_record_count_preserved_during_paging(self, m_debug):
        g = self.TG()
        assert g.record_count == 5
        assert m_debug.call_count == 3
        g.set_paging(25, 1)
        assert g.record_count == 5
        assert m_debug.call_count == 3

    @mock.patch('logging.Logger.debug')
    def test_record_count_cleared_during_filter(self, m_debug):
        g = self.KeyGrid()
        g.record_count  # noqa: B018
        assert m_debug.call_count == 3
        g.set_filter('id', 'eq', '5')
        g.record_count  # noqa: B018
        assert m_debug.call_count == 6

    def test_column_iterators_for_rendering(self):
        class TG(Grid):
            Column('C1', Person.firstname)
            Column('C1.5', Person.firstname.label('fn2'), render_in=None)
            Column('C2', Person.lastname, render_in='xlsx')
            BoolColumn('C3', Person.inactive, render_in=('xlsx', 'html'))
            YesNoColumn('C4', Person.inactive.label('yesno'), render_in='html')
            Column('C5', Person.firstname.label('fn3'), render_in='xlsx')
            Column('C6', Person.firstname.label('fn4'), render_in=('csv'))
            Column(
                'C7',
                Person.firstname.label('fn5'),
                render_in=('xlsx', 'html', 'csv'),
                visible=False,
            )

        tg = TG()

        html_cols = tuple(tg.iter_columns('html'))
        assert len(html_cols) == 3
        assert html_cols[0].key == 'firstname'
        assert html_cols[1].key == 'inactive'
        assert html_cols[2].key == 'yesno'

        xlsx_cols = tuple(tg.iter_columns('xlsx'))
        assert len(xlsx_cols) == 4
        assert xlsx_cols[0].key == 'firstname'
        assert xlsx_cols[1].key == 'lastname'
        assert xlsx_cols[2].key == 'inactive'
        assert xlsx_cols[3].key == 'fn3'
        csv_cols = tuple(tg.iter_columns('csv'))
        assert len(csv_cols) == 2
        assert csv_cols[1].key == 'fn4'

    def test_grid_inheritance(self):
        class SomeGrid(Grid):
            Column('n1', Person.firstname)

        class SomeGrid2(SomeGrid):
            Column('n2', Person.lastname)

        pg = SomeGrid2()
        assert len(pg.columns) == 2
        assert pg.columns[1].key == 'lastname'

    def test_export_as_response(self):
        export_xlsx = mock.MagicMock()

        class TG(Grid):
            Column('First Name', Person.firstname)

            def set_renderers(self):
                super().set_renderers()
                self.xlsx = export_xlsx

        grid = TG()
        grid.set_export_to('xlsx')
        grid.export_as_response()
        export_xlsx.as_response.assert_called_once_with(None, None)

        grid = TG()
        with pytest.raises(ValueError, match='No export format set'):
            grid.export_as_response()

    def test_export_as_response_with_csv(self):
        export_csv = mock.MagicMock()

        class TG(Grid):
            Column('First Name', Person.firstname)
            allowed_export_targets: ClassVar = {'csv': CSV}

            def set_renderers(self):
                super().set_renderers()
                self.csv = export_csv

        grid = TG()
        grid.set_export_to('csv')
        grid.export_as_response()
        export_csv.as_response.assert_called_once_with()

    def test_search_expressions_filtering_nones(self):
        class NonSearchingFilter(FilterBase):
            def get_search_expr(self):
                return None

        class CTG(Grid):
            Column('First Name', Person.firstname, NonSearchingFilter)

        assert len(CTG().search_expression_generators) == 0

    def test_search_expressions_generate_nones(self):
        class NonSearchingFilter(FilterBase):
            def get_search_expr(self):
                return lambda value: None

        class CTG(Grid):
            Column('First Name', Person.firstname, NonSearchingFilter)

        assert len(CTG().search_expression_generators) == 1
        g = CTG()
        g.search_value = 'foo'
        assert_not_in_query(g, 'WHERE')

    def test_search_expressions_uncallable(self):
        class BadFilter(FilterBase):
            def get_search_expr(self):
                return 'foo'

        class CTG(Grid):
            Column('First Name', Person.firstname, BadFilter)

        with pytest.raises(Exception, match='bad filter search expression: foo is not callable'):
            CTG().search_expression_generators  # noqa: B018

    def test_search_blank(self):
        class CTG(Grid):
            Column('First Name', Person.firstname, TextFilter)
            Column('Last Name', Person.lastname, TextFilter)

        g = CTG()
        g.search_value = ''
        assert_not_in_query(g.build_query(), 'WHERE')

    def test_search_query(self):
        class CTG(Grid):
            Column('First Name', Person.firstname, TextFilter)
            Column('Last Name', Person.lastname, TextFilter)

        g = CTG()
        g.search_value = 'foo'
        if db.engine.dialect.name == 'sqlite':
            search_where = (
                "WHERE lower(persons.firstname) LIKE lower('%foo%')"
                " OR lower(persons.last_name) LIKE lower('%foo%')"
            )
        elif db.engine.dialect.name == 'postgresql':
            search_where = (
                "WHERE persons.firstname ILIKE '%foo%' OR persons.last_name ILIKE '%foo%'"
            )
        elif db.engine.dialect.name == 'mssql':
            search_where = "WHERE persons.firstname LIKE '%foo%' OR persons.last_name LIKE '%foo%'"
        assert_in_query(g, search_where)

    def test_search_query_with_aggregate_column(self):
        class TG(Grid):
            Column('First Name', Person.firstname, TextFilter)
            Column('Count', sasql.func.count(Person.id).label('num_people'), AggregateIntFilter)

            # Column in set for filtering only. Not in select, so not needed in grouping.
            Column('Last Name', 'no_expr', TextFilter(Person.lastname), visible=False)

            def query_prep(self, query, has_sort, has_filters):
                return query.group_by(Person.firstname)

        g = TG()
        g.search_value = 'foo'

        search_expr = {
            'sqlite': (
                "WHERE lower(persons.firstname) LIKE lower('%foo%')"
                " OR lower(persons.last_name) LIKE lower('%foo%')"
            ),
            'postgresql': (
                "WHERE persons.firstname ILIKE '%foo%' OR persons.last_name ILIKE '%foo%'"
            ),
            'mssql': ("WHERE persons.firstname LIKE '%foo%' OR persons.last_name LIKE '%foo%'"),
        }[db.engine.dialect.name]
        assert_in_query(g, search_expr)
        g.records  # noqa: B018

        g.clear_record_cache()
        g.filtered_cols.pop('firstname')
        g.filtered_cols.pop('no_expr')
        search_expr = {
            'sqlite': ("HAVING CAST(count(persons.id) AS VARCHAR) LIKE '%foo%'"),
            'postgresql': ("HAVING CAST(count(persons.id) AS VARCHAR) LIKE '%foo%'"),
            'mssql': ("HAVING CAST(count(persons.id) AS NVARCHAR(max)) LIKE '%foo%'"),
        }[db.engine.dialect.name]
        assert_in_query(g, search_expr)
        g.records  # noqa: B018

    def test_column_keys_unique(self):
        grid = self.KeyGrid()
        assert grid.column('id').expr == Person.id
        assert grid.column('id_1').expr == Stopwatch.id

    def test_no_expression_column_key(self):
        """Test that columns without an expression get a key"""
        grid = self.KeyGrid()
        assert grid.column('no_expression')
        assert grid.column('no_expression_1')
        assert grid.column('unnamed_expression')

    def test_column_keys_unique_query_default_sort(self):
        grid = self.KeyGrid()
        grid.query_default_sort = 'id'
        assert_in_query(grid, 'ORDER BY stopwatches.id')
        grid.query_default_sort = Stopwatch.id
        assert_in_query(grid, 'ORDER BY stopwatches.id')
        grid.query_default_sort = Person.id
        assert_in_query(grid, 'ORDER BY persons.id')

    @_inrequest('/foo?op(id)=gte&v1(id)=0')
    def test_column_keys_unique_filter_persons(self):
        grid = self.KeyGrid()
        grid.apply_qs_args()
        assert_in_query(grid, 'WHERE persons.id >= 0')

    @_inrequest('/foo?op(id_1)=gte&v1(id_1)=0')
    def test_column_keys_unique_filter_stopwatches(self):
        grid = self.KeyGrid()
        grid.apply_qs_args()
        assert_in_query(grid, 'WHERE stopwatches.id >= 0')

    @_inrequest('/foo?sort1=id_1&sort2=id')
    def test_column_keys_unique_sort(self):
        grid = self.KeyGrid()
        grid.apply_qs_args()
        assert_in_query(grid, 'ORDER BY stopwatches.id, persons.id')

    def test_query_select_from(self):
        grid = self.KeyGrid()
        grid.query_joins = None
        grid.query_select_from = Stopwatch
        query = grid.build_query()
        assert len(query._from_obj) == 1
        assert query._from_obj[0] == Stopwatch.__table__


class TestQueryStringArgs:
    @classmethod
    def setup_class(cls):
        Status.delete_cascaded()
        Status.testing_create('pending')
        Status.testing_create('in process')
        Status.testing_create('complete')
        Person.testing_create('bob')
        Person.testing_create('bob')
        Person.testing_create()

    @_inrequest('/foo?dg_perpage=1&dg_onpage=2')
    @pytest.mark.parametrize('grid_cls', [PeopleGrid, PeopleGridByConfig])
    def test_qs_prefix(self, grid_cls):
        pg = grid_cls(qs_prefix='dg_')
        pg.apply_qs_args()
        assert pg.on_page == 2
        assert pg.per_page == 1

    @_inrequest('/foo?perpage=1&onpage=2')
    def test_qs_paging(self):
        pg = PeopleGrid()
        pg.apply_qs_args()
        assert pg.on_page == 2
        assert pg.per_page == 1

        # make sure the corret values get applied to the query
        if db.engine.dialect.name == 'mssql':
            assert_in_query(pg, 'OFFSET 1 ROWS')
            assert_in_query(pg, 'FETCH FIRST 1 ROWS ONLY')
        else:
            assert_in_query(pg, 'LIMIT 1 OFFSET 1')

        assert pg.build_qs_args() == 'onpage=2&perpage=1'

    @_inrequest('/foo?perpage=5&onpage=foo')
    def test_qs_onpage_invalid(self):
        pg = PeopleGrid()
        pg.apply_qs_args()
        assert pg.on_page == 1
        assert pg.per_page == 5
        assert pg.user_warnings[0] == '"onpage" grid argument invalid, ignoring'
        assert pg.build_qs_args() == 'perpage=5'

    @_inrequest('/foo?perpage=5&onpage=-1')
    def test_qs_onpage_negative(self):
        pg = PeopleGrid()
        pg.apply_qs_args()
        assert pg.on_page == 1
        assert pg.per_page == 5
        assert len(pg.user_warnings) == 0

    @_inrequest('/foo?perpage=1&onpage=100')
    def test_qs_onpage_too_high(self):
        pg = PeopleGrid()
        pg.apply_qs_args()
        assert pg.on_page == 3
        assert pg.per_page == 1
        assert len(pg.user_warnings) == 0

        # the records should be the same as if we were on the last page
        assert len(pg.records) == 1

    @_inrequest('/foo?perpage=foo&onpage=2')
    def test_qs_perpage_invalid(self):
        pg = PeopleGrid()
        pg.apply_qs_args()
        assert pg.on_page == 1
        assert pg.per_page == 50
        assert pg.user_warnings[0] == '"perpage" grid argument invalid, ignoring'

    def test_qs_paging_defaults(self):
        pg = PeopleGrid()
        pg.apply_qs_args(grid_args={'onpage': None, 'perpage': None})
        assert pg.on_page == 1
        assert pg.per_page == 50
        assert not pg.user_warnings
        assert pg.build_qs_args() == ''

    @_inrequest('/foo?perpage=-1&onpage=2')
    def test_qs_perpage_negative(self):
        pg = PeopleGrid()
        pg.apply_qs_args()
        assert pg.on_page == 2
        assert pg.per_page == 1
        assert len(pg.user_warnings) == 0

    @_inrequest('/foo?sort1=foo&sort2=firstname&sort3=-status')
    def test_qs_sorting(self):
        pg = PeopleGrid()
        pg.apply_qs_args()
        assert pg.order_by == [('firstname', False), ('status', True)]
        assert pg.user_warnings[0] == 'can\'t sort on invalid key "foo"'
        assert pg.build_qs_args() == 'sort1=firstname&sort2=-status'

    @_inrequest('/foo?sort1=fullname&sort2=firstname&sort3=-status')
    def test_qs_sorting_not_allowed(self):
        pg = PeopleGrid()
        pg.apply_qs_args()
        assert pg.order_by == [('firstname', False), ('status', True)]
        assert pg.user_warnings[0] == 'can\'t sort on invalid key "fullname"'

    @_inrequest('/foo?sort1=')
    def test_qs_sorting_ignores_emptystring(self):
        pg = PeopleGrid()
        pg.apply_qs_args()
        assert pg.order_by == []
        assert len(pg.user_warnings) == 0

    def test_qs_filtering(self):
        first_id = Status.query.filter_by(label='pending').one().id
        second_id = Status.query.filter_by(label='in process').one().id
        pg = PeopleGrid()
        with pg.manager.test_request_context(
            '/foo?op(firstname)=eq&v1(firstname)=fn001&op(status)=is'
            f'&v1(status)={first_id}&v1(status)={second_id}',
        ):
            pg.apply_qs_args()
        assert pg.columns[0].filter.op == 'eq'
        assert pg.columns[0].filter.value1 == 'fn001'
        assert pg.columns[0].filter.value2 is None

        assert pg.columns[4].filter.op == 'is'
        assert pg.columns[4].filter.value1 == [first_id, second_id]
        assert pg.columns[4].filter.value2 is None

        assert pg.build_qs_args() == (
            'op(firstname)=eq&op(status)=is&v1(firstname)=fn001'
            f'&v1(status)={first_id}&v1(status)={second_id}'
        )

    @_inrequest('/foo')
    def test_qs_filtering_default_op(self):
        from webgrid_ta.grids import DefaultOpGrid

        pg = DefaultOpGrid()
        pg.apply_qs_args()
        assert pg.columns[0].filter.op == 'eq'
        assert pg.columns[0].filter.value1 is None

    @_inrequest('/foo?op(firstname)=!eq&v1(firstname)=bob')
    def test_qs_filtering_default_op_override(self):
        from webgrid_ta.grids import DefaultOpGrid

        pg = DefaultOpGrid()
        pg.apply_qs_args()
        assert pg.columns[0].filter.op == '!eq'
        assert pg.columns[0].filter.value1 == 'bob'

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob&perpage=1&onpage=100')
    def test_qs_paging_doesnt_get_page_count_before_filters_are_handled(self):
        pg = PeopleGrid()
        pg.apply_qs_args()
        # this ensures that the filtering in apply_qs_args() takes place before
        # the paging.  Since the paging section uses .page_count, we have to
        # make sure the filters get processed first.  Otherwise an incorrect
        # page count gets cached.
        assert pg.record_count == 2

    @_inrequest('/foo')
    def test_qs_no_session(self):
        pg = PeopleGrid()
        pg.apply_qs_args()
        assert pg.column('firstname').filter.op is None

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob&perpage=1&onpage=100')
    def test_qs_keyed_session(self):
        pg = PeopleGrid()
        pg.apply_qs_args()
        flask.request.args['op(firstname)'] = '!eq'
        pg2 = PeopleGrid()
        pg2.apply_qs_args()
        assert pg2.column('firstname').filter.op == '!eq'
        flask.request.args = MultiDict([('session_key', pg.session_key)])
        pg = PeopleGrid()
        pg.apply_qs_args()
        assert pg.column('firstname').filter.op == 'eq'

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob&perpage=1&onpage=100')
    def test_grid_args_ignores_url(self):
        pg = PeopleGrid()
        pg.apply_qs_args(
            grid_args=MultiDict(
                {
                    'op(firstname)': 'contains',
                    'v1(firstname)': 'bill',
                },
            ),
        )
        assert pg.column('firstname').filter.op == 'contains'
        assert pg.column('firstname').filter.value1 == 'bill'

    @_inrequest('/foo?op(firstname)=&v1(firstname)=foo&op(status)=&v1(status)=1')
    def test_qs_blank_operator(self):
        pg = PeopleGrid()
        pg.apply_qs_args()
        assert pg.columns[0].filter.op is None
        assert pg.columns[0].filter.value1 is None
        assert pg.columns[0].filter.value2 is None

        assert pg.columns[4].filter.op is None
        assert pg.columns[4].filter.value1 is None
        assert pg.columns[4].filter.value2 is None

    @_inrequest('/foo?sort1=legacycol1&sort2=legacycol2')
    def test_sa_expr_sort(self):
        class AGrid(Grid):
            Column('First Name', 'firstname')
            Column('LC1', Person.legacycol1)
            Column('LC2', Person.legacycol2)

            def query_base(self, has_sort, has_filters):
                query = db.session.query(
                    Person.id,
                    Person.firstname,
                    Person.legacycol1,
                    Person.legacycol2,
                )
                return query

        g = AGrid()
        g.apply_qs_args()

        # this will fail if we are not using SA expressions correctly to sort
        g.records  # noqa: B018

    @_inrequest('/thepage?export_to=xlsx')
    def test_export_to_xlsx(self):
        g = PeopleGrid()
        g.apply_qs_args()
        assert g.export_to == 'xlsx'

    @_inrequest('/thepage?export_to=foo')
    def test_export_to_unrecognized(self):
        g = PeopleGrid()
        g.apply_qs_args()
        assert g.export_to is None

    @_inrequest('/foo?op(id)=eq&v1(id)=d')
    def test_exc_msg_in_warnings(self):
        class TGrid(Grid):
            Column('T', Person.id, IntFilter)

        g = TGrid()
        g.apply_qs_args()
        assert g.user_warnings[0] == 'T: Please enter an integer value.'

    @_inrequest('/foo?search=bar')
    def test_qs_search(self):
        g = PeopleGrid()
        g.enable_search = True
        g.apply_qs_args()
        assert g.search_value == 'bar'

    @_inrequest('/foo?search=bar')
    def test_qs_search_disabled(self):
        g = PeopleGrid()
        g.enable_search = False
        g.apply_qs_args()
        assert g.search_value is None

    @_inrequest('/foo?search=bar')
    def test_qs_no_searchable_columns(self):
        class TG(Grid):
            Column('First Name', Person.firstname)

        g = TG()
        g.enable_search = True
        g.apply_qs_args()
        assert g.search_value is None


class GridPrefixTestBase:
    def test_passthru(self):
        source_args = MultiDict([('foo', 'bar'), ('baz', 'bin')])
        source_args_copy = source_args.copy()
        grid = PeopleGrid()
        setattr(grid.manager, self.manager_arg_method, MagicMock(return_value=source_args))
        loader = self.loader_cls(grid.manager)
        result = loader.get_args(grid, MultiDict())
        assert result == source_args == source_args_copy

    def test_merge_with_previous(self):
        source_args = MultiDict([('foo', 'bar'), ('baz', 'bin')])
        previous_args = MultiDict([('foo', 'fog'), ('bong', 'bing')])
        grid = PeopleGrid()
        setattr(grid.manager, self.manager_arg_method, MagicMock(return_value=source_args))
        loader = self.loader_cls(grid.manager)
        result = loader.get_args(grid, previous_args)
        assert result == MultiDict(
            [('foo', 'bar'), ('foo', 'fog'), ('baz', 'bin'), ('bong', 'bing')],
        )

    def test_qs_prefix_filter(self):
        source_args = MultiDict([('foo', 'bar'), ('baz', 'bin'), ('boo', 'hoo'), ('baz', 'bid')])
        source_args_copy = source_args.copy()
        grid = PeopleGrid(qs_prefix='b')
        setattr(grid.manager, self.manager_arg_method, MagicMock(return_value=source_args))
        loader = self.loader_cls(grid.manager)
        result = loader.get_args(grid, MultiDict())
        assert source_args == source_args_copy
        assert result == MultiDict([('az', 'bin'), ('az', 'bid'), ('oo', 'hoo')])

    def test_reset(self):
        source_args = MultiDict([('foo', 'bar'), ('baz', 'bin'), ('dgreset', '1')])
        grid = PeopleGrid()
        setattr(grid.manager, self.manager_arg_method, MagicMock(return_value=source_args))
        loader = self.loader_cls(grid.manager)
        result = loader.get_args(grid, MultiDict())
        assert result == MultiDict([('dgreset', 1)])

    def test_reset_with_session_key(self):
        source_args = MultiDict(
            [('foo', 'bar'), ('baz', 'bin'), ('dgreset', '1'), ('session_key', '123')],
        )
        grid = PeopleGrid()
        setattr(grid.manager, self.manager_arg_method, MagicMock(return_value=source_args))
        loader = self.loader_cls(grid.manager)
        result = loader.get_args(grid, MultiDict())
        assert result == MultiDict([('dgreset', 1), ('session_key', '123')])


class TestRequestFormLoader(GridPrefixTestBase):
    loader_cls = RequestFormLoader
    manager_arg_method = 'request_form_args'


class TestRequestArgsLoader(GridPrefixTestBase):
    loader_cls = RequestArgsLoader
    manager_arg_method = 'request_url_args'


class TestRequestJsonLoader:
    def ok_values(self):
        return {
            'search_expr': 'foo',
            'filters': {
                'test': {'op': 'eq', 'value1': 'toast', 'value2': 'taft'},
                'test2': {'op': 'in', 'value1': 'tarp', 'value2': None},
            },
            'paging': {'on_page': 2, 'per_page': 20},
            'sort': [{'key': 'bar', 'flag_desc': False}, {'key': 'baz', 'flag_desc': True}],
        }

    def test_load(self):
        data = self.ok_values()
        grid = PeopleGrid()
        grid.manager.request_json = MagicMock(return_value=data)
        loader = RequestJsonLoader(grid.manager)
        result = loader.get_args(grid, MultiDict())
        assert result == MultiDict(
            [
                ('search', 'foo'),
                ('op(test)', 'eq'),
                ('v1(test)', 'toast'),
                ('v2(test)', 'taft'),
                ('op(test2)', 'in'),
                ('v1(test2)', 'tarp'),
                ('onpage', 2),
                ('perpage', 20),
                ('sort1', 'bar'),
                ('sort2', '-baz'),
                ('export_to', None),
            ],
        )

    def test_merge_with_previous(self):
        data = self.ok_values()
        grid = PeopleGrid()
        grid.manager.request_json = MagicMock(return_value=data)
        loader = RequestJsonLoader(grid.manager)
        result = loader.get_args(
            grid,
            MultiDict(
                [
                    ('search', 'oof'),
                    ('onpage', 1),
                    ('sort3', 'bong'),
                ],
            ),
        )
        assert result == MultiDict(
            [
                ('search', 'foo'),
                ('search', 'oof'),
                ('op(test)', 'eq'),
                ('v1(test)', 'toast'),
                ('v2(test)', 'taft'),
                ('op(test2)', 'in'),
                ('v1(test2)', 'tarp'),
                ('onpage', 2),
                ('onpage', 1),
                ('perpage', 20),
                ('sort1', 'bar'),
                ('sort2', '-baz'),
                ('sort3', 'bong'),
                ('export_to', None),
            ],
        )

    def test_load_empty(self):
        grid = PeopleGrid()
        grid.manager.request_json = MagicMock(return_value=None)
        loader = RequestJsonLoader(grid.manager)
        result = loader.get_args(grid, MultiDict())
        assert result == MultiDict()


class TestWebSessionArgsLoader:
    @_inrequest('/foo')
    def test_no_session_present(self):
        source_args = MultiDict([('foo', 'bar'), ('baz', 'bin')])
        source_args_copy = source_args.copy()
        grid = PeopleGrid()
        loader = WebSessionArgsLoader(grid.manager)
        result = loader.get_args(grid, source_args)
        assert source_args == source_args_copy
        assert result == MultiDict(
            [
                ('foo', 'bar'),
                ('baz', 'bin'),
                ('__foreign_session_loaded__', False),
            ],
        )

    @_inrequest('/foo')
    def test_empty_session_not_stored(self):
        source_args = MultiDict([])
        grid = PeopleGrid()
        loader = WebSessionArgsLoader(grid.manager)
        loader.get_args(grid, source_args)
        assert '_PeopleGrid' not in flask.session['dgsessions']
        assert grid.session_key not in flask.session['dgsessions']

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob')
    def test_matching_session_not_stored(self):
        pg = PeopleGrid()
        loader = WebSessionArgsLoader(pg.manager)
        loader.get_args(pg, flask.request.args)
        pg2 = PeopleGrid()
        loader.get_args(pg2, flask.request.args)
        assert '_PeopleGrid' in flask.session['dgsessions']
        assert pg.session_key in flask.session['dgsessions']
        assert pg2.session_key not in flask.session['dgsessions']

    @_inrequest('/foo?dgreset=1&op(firstname)=eq&v1(firstname)=bob')
    def test_reset_result(self):
        pg = PeopleGrid()
        loader = WebSessionArgsLoader(pg.manager)
        args = loader.get_args(pg, flask.request.args)
        assert args == MultiDict([('dgreset', 1), ('session_key', None)])

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob&perpage=1&onpage=100')
    def test_reset_removes_session(self):
        pg = PeopleGrid()
        loader = WebSessionArgsLoader(pg.manager)
        loader.get_args(pg, flask.request.args)
        flask.request.args = MultiDict(
            [
                ('session_key', pg.session_key),
                ('dgreset', '1'),
            ],
        )
        pg = PeopleGrid()
        loader.get_args(pg, flask.request.args)
        assert '_PeopleGrid' not in flask.session['dgsessions']
        assert pg.session_key not in flask.session['dgsessions']

    @_inrequest('/foo')
    def test_session_load_from_none(self):
        # test backwards compatibility for multidict load
        flask.session['dgsessions'] = {}
        grid = PeopleGrid()
        loader = WebSessionArgsLoader(grid.manager)
        args = loader.get_session_store(grid, MultiDict())
        assert args == MultiDict([])

    @_inrequest('/foo')
    def test_session_load_from_multidict(self):
        # test backwards compatibility for multidict load
        flask.session['dgsessions'] = {'_PeopleGrid': MultiDict([('a', 'b'), ('a', 'c')])}
        grid = PeopleGrid()
        loader = WebSessionArgsLoader(grid.manager)
        args = loader.get_session_store(grid, MultiDict())
        assert args == MultiDict([('a', 'b'), ('a', 'c')])

    @_inrequest('/foo')
    def test_session_load_from_dict(self):
        # test backwards compatibility for dict load
        flask.session['dgsessions'] = {'_PeopleGrid': {'a': 'b', 'c': 'd'}}
        grid = PeopleGrid()
        loader = WebSessionArgsLoader(grid.manager)
        args = loader.get_session_store(grid, MultiDict())
        assert args == MultiDict([('a', 'b'), ('c', 'd')])

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob&perpage=1&onpage=100')
    def test_default_session(self):
        pg = PeopleGrid()
        loader = WebSessionArgsLoader(pg.manager)
        loader.get_args(pg, flask.request.args)
        pg2 = PeopleGrid()
        loader = WebSessionArgsLoader(pg2.manager)
        args = loader.get_args(pg2, MultiDict())
        assert args['op(firstname)'] == 'eq'

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob&perpage=1&onpage=100')
    def test_keyed_session(self):
        pg = PeopleGrid()
        loader = WebSessionArgsLoader(pg.manager)
        loader.get_args(pg, flask.request.args)
        flask.request.args['op(firstname)'] = '!eq'
        pg2 = PeopleGrid()
        args = loader.get_args(pg2, flask.request.args)
        assert args['op(firstname)'] == '!eq'
        flask.request.args = MultiDict([('session_key', pg.session_key)])
        pg = PeopleGrid()
        args = loader.get_args(pg, flask.request.args)
        assert args['op(firstname)'] == 'eq'

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob&perpage=1&onpage=100&sort1=foo')
    def test_page_args_always_applied(self):
        pg = PeopleGrid()
        loader = WebSessionArgsLoader(pg.manager)
        loader.get_args(pg, flask.request.args)
        flask.request.args = MultiDict(
            [
                ('session_key', pg.session_key),
                ('perpage', '15'),
                ('onpage', '100'),
            ],
        )
        pg = PeopleGrid()
        args = loader.get_args(pg, flask.request.args)
        assert args['op(firstname)'] == 'eq'
        assert args['perpage'] == '15'
        assert args['onpage'] == '100'
        assert args['sort1'] == 'foo'
        assert '["perpage", "15"]' in flask.session['dgsessions'][pg.session_key]

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob&perpage=1&onpage=100&sort1=foo')
    def test_sort_args_always_applied(self):
        pg = PeopleGrid()
        loader = WebSessionArgsLoader(pg.manager)
        loader.get_args(pg, flask.request.args)
        flask.request.args = MultiDict(
            [
                ('session_key', pg.session_key),
                ('sort1', 'bar'),
            ],
        )
        pg = PeopleGrid()
        args = loader.get_args(pg, flask.request.args)
        assert args['op(firstname)'] == 'eq'
        assert args['perpage'] == '1'
        assert args['sort1'] == 'bar'
        assert '["sort1", "bar"]' in flask.session['dgsessions'][pg.session_key]

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob&perpage=1&onpage=100')
    def test_keyed_session_with_export(self):
        pg = PeopleGrid()
        loader = WebSessionArgsLoader(pg.manager)
        loader.get_args(pg, flask.request.args)
        flask.request.args['op(firstname)'] = '!eq'
        pg2 = PeopleGrid()
        args = loader.get_args(pg2, flask.request.args)
        assert args['op(firstname)'] == '!eq'
        flask.request.args = MultiDict(
            [
                ('session_key', pg.session_key),
                ('export_to', 'xls'),
            ],
        )
        pg = PeopleGrid()
        args = loader.get_args(pg, flask.request.args)
        assert args['op(firstname)'] == 'eq'
        assert args['export_to'] == 'xls'
        assert 'export_to' not in flask.session['dgsessions'][pg.session_key]

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob&perpage=1&onpage=100')
    def test_keyed_session_with_override(self):
        pg = PeopleGrid()
        loader = WebSessionArgsLoader(pg.manager)
        loader.get_args(pg, flask.request.args)
        flask.request.args = MultiDict(
            [
                ('session_key', pg.session_key),
                ('session_override', 1),
                ('op(createdts)', '!eq'),
                ('v1(createdts)', '2017-05-06'),
            ],
        )
        pg2 = PeopleGrid()
        args = loader.get_args(pg2, flask.request.args)
        assert args['op(firstname)'] == 'eq'
        assert args['v1(firstname)'] == 'bob'
        assert args['op(createdts)'] == '!eq'
        assert args['v1(createdts)'] == '2017-05-06'

    @_inrequest('/foo?op(account_type)=is&v1(account_type)=admin')
    def test_keyed_session_with_override_multifilter(self):
        pg = PeopleGrid()
        loader = WebSessionArgsLoader(pg.manager)
        loader.get_args(pg, flask.request.args)
        flask.request.args = MultiDict(
            [
                ('session_key', pg.session_key),
                ('session_override', 1),
                ('op(account_type)', 'is'),
                ('v1(account_type)', 'manager'),
            ],
        )
        pg2 = PeopleGrid()
        args = loader.get_args(pg2, flask.request.args)
        assert args['op(account_type)'] == 'is'
        assert args.getlist('v1(account_type)') == ['manager']

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob&perpage=1&onpage=100')
    def test_keyed_session_without_override(self):
        pg = PeopleGrid()
        loader = WebSessionArgsLoader(pg.manager)
        loader.get_args(pg, flask.request.args)
        flask.request.args = MultiDict(
            [
                ('session_key', pg.session_key),
                ('op(createdts)', '!eq'),
                ('v1(createdts)', '2017-05-06'),
            ],
        )
        pg2 = PeopleGrid()
        args = loader.get_args(pg2, flask.request.args)
        assert 'op(firstname)' not in args
        assert 'v1(firstname)' not in args
        assert args['op(createdts)'] == '!eq'
        assert args['v1(createdts)'] == '2017-05-06'

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob&perpage=1&onpage=100')
    def test_apply_prevents_session_load(self):
        pg = PeopleGrid()
        loader = WebSessionArgsLoader(pg.manager)
        loader.get_args(pg, flask.request.args)
        flask.request.args = MultiDict(
            [
                ('session_key', pg.session_key),
                ('op(createdts)', 'today'),
                ('apply', None),
            ],
        )
        pg2 = PeopleGrid()
        args = loader.get_args(pg2, flask.request.args)
        assert 'op(firstname)' not in args
        assert 'v1(firstname)' not in args
        assert 'apply' not in flask.session['dgsessions'][pg.session_key]

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob&perpage=1&onpage=100')
    def test_expired_session_cleaned_up(self):
        with mock.patch(
            'webgrid.extensions.arrow.utcnow',
            lambda *args: arrow.get('2021-01-05 15:14:13'),
        ):
            pg = PeopleGrid()
            loader = WebSessionArgsLoader(pg.manager)
            loader.get_args(pg, flask.request.args)

        assert '_PeopleGrid' in flask.session['dgsessions']
        assert pg.session_key in flask.session['dgsessions']

        with mock.patch(
            'webgrid.extensions.arrow.utcnow',
            lambda *args: arrow.get('2021-01-06 03:14:15'),
        ):
            pg = PeopleGrid()
            loader = WebSessionArgsLoader(pg.manager)
            loader.get_args(pg, MultiDict([]))

        assert '_PeopleGrid' not in flask.session['dgsessions']
        assert pg.session_key not in flask.session['dgsessions']

    @_inrequest('/foo')
    def test_expired_session_no_stamp(self):
        flask.session['dgsessions'] = {'_PeopleGrid': MultiDict([('a', 'b'), ('a', 'c')])}

        pg = PeopleGrid()
        loader = WebSessionArgsLoader(pg.manager)
        loader.cleanup_expired_sessions()
        assert flask.session['dgsessions']['_PeopleGrid'] == MultiDict([('a', 'b'), ('a', 'c')])

    @_inrequest('/foo?op(firstname)=eq&v1(firstname)=bob&perpage=1&onpage=100')
    def test_expired_session_no_max_hours(self):
        with mock.patch(
            'webgrid.extensions.arrow.utcnow',
            lambda *args: arrow.get('2021-01-05 15:14:13'),
        ):
            pg = PeopleGrid()
            loader = WebSessionArgsLoader(pg.manager)
            loader.get_args(pg, flask.request.args)

        assert '_PeopleGrid' in flask.session['dgsessions']
        assert pg.session_key in flask.session['dgsessions']

        pg = PeopleGrid()
        pg.manager.session_max_hours = None
        loader = WebSessionArgsLoader(pg.manager)
        loader.get_args(pg, MultiDict([]))

        assert '_PeopleGrid' in flask.session['dgsessions']


class TestCustomJSONEncoder:
    def dump(self, value):
        return json.dumps(value, cls=CustomJsonEncoder)

    def test_date(self):
        assert self.dump(dt.date(2022, 5, 31)) == '"2022-05-31"'

    def test_arrow(self):
        assert self.dump(arrow.get(2022, 5, 31)) == '"2022-05-31T00:00:00+00:00"'

    def test_decimal(self):
        assert self.dump(Decimal('5')) == '5.0'

    def test_custom_unhandled(self):
        class Foo:
            def __str__(self):
                return 'I am awesome'

        assert self.dump(Foo()) == '"I am awesome"'



import logging
import urllib

import click
import flask

from webgrid_ta.app import create_app
from webgrid_ta.extensions import lazy_gettext as _
import webgrid_ta.model as model
from webgrid_ta.model.helpers import clear_db


log = logging.getLogger(__name__)

app = None


@click.group()
def main():
    """Run the Webgrid Test App CLI."""
    app = create_app('Dev')
    flask.ctx.AppContext(app).push()


@main.command('create-db')
@click.option('--clear', default=False, is_flag=True, help=_('DROP all DB objects first'))
def database_init(clear):
    if clear:
        clear_db()
        print(_('- db cleared'))

    model.load_db()
    print(_('- db loaded'))


@main.command('list-routes')
def list_routes():
    output = []
    for rule in flask.current_app.url_map.iter_rules():
        methods = ','.join(rule.methods)
        line = urllib.parse.unquote(f'{rule.endpoint:50s} {methods:20s} {rule}')
        output.append(line)

    for line in sorted(output):
        print(line)


main.add_command(flask.cli.run_command)


@main.command('verify-translations')
def verify_translations():
    from pathlib import Path

    from morphi.messages.validation import check_translations

    root_path = Path(__file__).resolve().parent.parent.parent
    check_translations(
        root_path,
        'webgrid',
    )


if __name__ == '__main__':
    main()



from typing import ClassVar

from webgrid import Column, DateColumn, DateTimeColumn, LinkColumnBase, NumericColumn, YesNoColumn
from webgrid.blazeweb import Grid
from webgrid.filters import DateTimeFilter, Operator, OptionsFilterBase, TextFilter, ops
from webgrid.renderers import CSV, XLS, XLSX
from webgrid_blazeweb_ta.model.orm import Person, Status


class FirstNameColumn(LinkColumnBase):
    def create_url(self, record):
        return f'/person-edit/{record.id}'


class FullNameColumn(LinkColumnBase):
    def extract_data(self, record):
        return f'{record.firstname} {record.lastname}'

    def create_url(self, record):
        return f'/person-edit/{record.id}'


class EmailsColumn(Column):
    def extract_data(self, recordset):
        return ', '.join([e.email for e in recordset.Person.emails])


class StatusFilter(OptionsFilterBase):
    operators = (
        Operator('o', 'open', None),
        ops.is_,
        ops.not_is,
        Operator('c', 'closed', None),
        ops.empty,
        ops.not_empty,
    )
    options_from = Status.pairs


class PeopleGrid(Grid):
    session_on = True
    allowed_export_targets: ClassVar = {'csv': CSV, 'xls': XLS, 'xlsx': XLSX}
    FirstNameColumn('First Name', Person.firstname, TextFilter)
    FullNameColumn('Full Name')
    YesNoColumn('Active', Person.inactive, reverse=True)
    EmailsColumn('Emails')
    Column('Status', Status.label.label('status'), StatusFilter(Status.id))
    DateTimeColumn('Created', Person.createdts, DateTimeFilter)
    DateColumn('Due Date', 'due_date')
    NumericColumn('Number', Person.numericcol, has_subtotal=True)

    def query_prep(self, query, has_sort, has_filters):
        query = (
            query.add_columns(Person.id, Person.lastname, Person.due_date)
            .add_entity(Person)
            .outerjoin(Person.status)
        )

        # default sort
        if not has_sort:
            query = query.order_by(Person.id)

        return query



from os import path
import warnings

from blazeweb.config import DefaultSettings


basedir = path.dirname(path.dirname(__file__))  # noqa: PTH120
app_package = path.basename(basedir)  # noqa: PTH119


class Default(DefaultSettings):
    def init(self):
        self.dirs.base = basedir
        self.app_package = app_package
        DefaultSettings.init(self)

        self.init_routing()

        self.add_component(app_package, 'sqlalchemy', 'sqlalchemybwc')
        self.add_component(app_package, 'webgrid', 'webgrid')
        self.add_component(app_package, 'templating', 'templatingbwc')

        self.name.full = 'WebGrid'
        self.name.short = 'WebGrid'

        # ignore the really big SA warning about lossy numeric types & sqlite
        warnings.filterwarnings('ignore', '.*support Decimal objects natively.*')

    def init_routing(self):
        self.add_route('/people/manage', endpoint='ManagePeople')


class Dev(Default):
    def init(self):
        Default.init(self)
        self.apply_dev_settings()

        self.db.url = 'sqlite://'

        # uncomment this if you want to use a database you can inspect
        # from os import path
        # self.db.url = 'sqlite:///%s' % path.join(self.dirs.data, 'test_application.db')


class Test(Default):
    def init(self):
        Default.init(self)
        self.apply_test_settings()

        self.db.url = 'sqlite://'

        # uncomment this if you want to use a database you can inspect
        # from os import path
        # self.db.url = 'sqlite:///%s' % path.join(self.dirs.data, 'test_application.db')


try:
    from site_settings import *  # noqa
except ImportError as e:
    if "No module named 'site_settings'" not in str(e):
        raise



from blazeweb.application import WSGIApp
from blazeweb.middleware import full_wsgi_stack
from blazeweb.scripting import application_entry
from sqlalchemybwc.lib.middleware import SQLAlchemyApp

import webgrid_blazeweb_ta.config.settings as settingsmod


def make_wsgi(profile='Dev'):
    app = WSGIApp(settingsmod, profile)

    app = SQLAlchemyApp(app)

    return full_wsgi_stack(app)


def script_entry():
    application_entry(make_wsgi)


if __name__ == '__main__':
    script_entry()



from pathlib import Path

from nox import Session, options, parametrize, session


package_path = Path.cwd()
tests_dpath = package_path / 'tests'
docs_dpath = package_path / 'docs'

py_all = ['3.10', '3.11', '3.12', '3.13']
py_single = py_all[-1:]
py_311 = ['3.11']

options.default_venv_backend = 'uv'


def pytest_run(session: Session, *args, **env):
    session.run(
        'pytest',
        '-ra',
        '--tb=native',
        '--strict-markers',
        '--cov',
        '--cov-config=.coveragerc',
        f'--cov-report=xml:{package_path}/ci/coverage/{session.name}.xml',
        '--no-cov-on-fail',
        'tests/webgrid_tests',
        *args,
        *session.posargs,
        env=env,
    )


def uv_sync(session: Session, *groups, project, extra=None):
    project_args = () if project else ('--no-install-project',)
    group_args = [arg for group in groups for arg in ('--group', group)]
    extra_args = ('--extra', extra) if extra else ()
    run_args = (
        'uv',
        'sync',
        '--active',
        '--no-default-groups',
        *project_args,
        *group_args,
        *extra_args,
    )
    session.run(*run_args)


@session(py=py_all)
@parametrize('db', ['pg', 'sqlite'])
def pytest(session: Session, db: str):
    uv_sync(session, 'tests', project=True)
    pytest_run(session, WEBTEST_DB=db)


@session(py=py_single)
def pytest_mssql(session: Session):
    uv_sync(session, 'tests', 'mssql', project=True)
    pytest_run(session, WEBTEST_DB='mssql')


@session(py=py_single)
def pytest_i18n(session: Session):
    uv_sync(session, 'tests', project=True, extra='i18n')
    pytest_run(session, WEBTEST_DB='sqlite')


@session(py=py_single)
def wheel(session: Session):
    """
    Package the wheel, install in the venv, and then run the tests for one version of Python.
    Helps ensure nothing is wrong with how we package the wheel.
    """
    uv_sync(session, 'tests', project=False)

    session.install('hatch', 'check-wheel-contents')
    version = session.run('hatch', 'version', silent=True, stderr=None).strip()
    wheel_fpath = package_path / 'tmp' / 'dist' / f'webgrid-{version}-py3-none-any.whl'

    if wheel_fpath.exists():
        wheel_fpath.unlink()

    session.run('hatch', 'build', '--clean')
    session.run('check-wheel-contents', wheel_fpath)
    session.run('uv', 'pip', 'install', wheel_fpath)

    out = session.run('python', '-c', 'import webgrid; print(webgrid.__file__)', silent=True)
    assert 'site-packages/webgrid/__init__.py' in out

    pytest_run(session, WEBTEST_DB='sqlite')


@session(py=py_single)
def precommit(session: Session):
    uv_sync(session, 'pre-commit', project=False)
    session.run(
        'pre-commit',
        'run',
        '--all-files',
    )


# Python 3.11 is required due to: https://github.com/level12/morphi/issues/11
@session(python=py_311)
def translations(session: Session):
    uv_sync(session, 'tests', project=True, extra='i18n')
    # This is currently failing due to missing translations
    # https://github.com/level12/webgrid/issues/194
    session.run(
        'python',
        'tests/webgrid_ta/manage.py',
        'verify-translations',
        env={'PYTHONPATH': tests_dpath},
    )


@session(py=py_single)
def docs(session: Session):
    uv_sync(session, 'tests', 'docs', project=True)
    session.run('make', '-C', docs_dpath, 'html', external=True)



from blazeutils.strings import randchars
import sqlalchemy as sa
import sqlalchemy.orm as saorm
from sqlalchemybwc import db
from sqlalchemybwc.lib.declarative import DefaultMixin, declarative_base


Base = declarative_base()


class Person(Base, DefaultMixin):
    __tablename__ = 'persons'

    id = sa.Column(sa.Integer, primary_key=True)
    firstname = sa.Column(sa.String(50))
    lastname = sa.Column('last_name', sa.String(50))
    inactive = sa.Column(sa.SmallInteger)
    state = sa.Column(sa.String(50))
    status_id = sa.Column(sa.Integer, sa.ForeignKey('statuses.id'))
    address = sa.Column(sa.Integer)
    createdts = sa.Column(sa.DateTime)
    sortorder = sa.Column(sa.Integer)
    floatcol = sa.Column(sa.Float)
    numericcol = sa.Column(sa.Numeric)
    boolcol = sa.Column(sa.Boolean)
    due_date = sa.Column(sa.Date)
    legacycol1 = sa.Column('LegacyColumn1', sa.String(50), key='legacycolumn')
    legacycol2 = sa.Column('LegacyColumn2', sa.String(50))

    status = saorm.relationship('Status')

    def __repr__(self):
        return f'<Person: "{self.id}, created: {self.createdts}">'

    @classmethod
    def testing_create(cls, firstname=None):
        firstname = firstname or randchars()
        return cls.add(firstname=firstname)

    @classmethod
    def delete_cascaded(cls):
        Email.delete_all()
        cls.delete_all()


class Email(Base, DefaultMixin):
    __tablename__ = 'emails'

    id = sa.Column(sa.Integer, primary_key=True)
    person_id = sa.Column(sa.Integer, sa.ForeignKey(Person.id), nullable=False)
    email = sa.Column(sa.String(50), nullable=False)

    person = saorm.relationship(Person, backref='emails')


class Status(Base, DefaultMixin):
    __tablename__ = 'statuses'

    id = sa.Column(sa.Integer, primary_key=True)
    label = sa.Column(sa.String(50), nullable=False, unique=True)
    flag_closed = sa.Column(sa.Integer, default=0)

    @classmethod
    def pairs(cls):
        return db.sess.query(cls.id, cls.label).order_by(cls.label)

    @classmethod
    def delete_cascaded(cls):
        Person.delete_cascaded()
        cls.delete_all()

    @classmethod
    def testing_create(cls, label=None):
        label = label or randchars()
        return cls.add(label=label)



"""
A collection of utilities for testing webgrid functionality in client applications
"""

import re
from unittest import mock
import urllib


try:
    import openpyxl
except ImportError:
    openpyxl = None
from pyquery import PyQuery
import sqlalchemy


def compiler_instance_factory(compiler, dialect, statement):
    class LiteralCompiler(compiler.__class__):
        def render_literal_value(self, value, type_):
            import datetime

            """
            For date and datetime values, convert to a string
            format acceptable to the dialect. That seems to be the
            so-called ODBC canonical date format which looks
            like this:

                yyyy-mm-dd hh:mi:ss.mmm(24h)

            For other data types, call the base class implementation.
            """
            if isinstance(value, datetime.datetime):
                return "'" + value.strftime('%Y-%m-%d %H:%M:%S.%f') + "'"
            elif isinstance(value, datetime.date):
                return "'" + value.strftime('%Y-%m-%d') + "'"
            elif isinstance(value, datetime.time):
                return f"'{value:%H:%M:%S.%f}'"
            elif isinstance(value, datetime.timedelta):
                return str(value)
            elif isinstance(value, str):
                return f"'{value}'"
            elif isinstance(value, list) and isinstance(type_, sqlalchemy.ARRAY):
                elements = [
                    self.render_literal_value(list_val, type_.item_type) for list_val in value
                ]
                return f'({", ".join(elements)})'
            elif value is None:
                return 'NULL'
            else:
                # Turn off double percent escaping, since we don't run these strings and
                # it creates a large number of differences for test cases
                with mock.patch.object(dialect.identifier_preparer, '_double_percents', False):
                    return super().render_literal_value(value, type_)

        def visit_bindparam(
            self,
            bindparam,
            within_columns_clause=False,
            literal_binds=False,
            **kwargs,
        ):
            return super().render_literal_bindparam(
                bindparam,
                within_columns_clause=within_columns_clause,
                literal_binds=literal_binds,
                **kwargs,
            )

        def visit_table(
            self,
            table,
            asfrom=False,
            iscrud=False,
            ashint=False,
            fromhints=None,
            use_schema=True,
            **kwargs,
        ):
            """Strip the default schema from table names when it is not needed"""
            ret_val = super().visit_table(
                table,
                asfrom,
                iscrud,
                ashint,
                fromhints,
                use_schema,
                **kwargs,
            )
            if dialect.name == 'mssql' and ret_val.startswith('dbo.'):
                return ret_val[4:]
            return ret_val

        def visit_column(self, column, add_to_result_map=None, include_table=True, **kwargs):
            """Strip the default schema from table names when it is not needed"""
            ret_val = super().visit_column(column, add_to_result_map, include_table, **kwargs)
            if dialect.name == 'mssql' and ret_val.startswith('dbo.'):
                return ret_val[4:]
            return ret_val

    return LiteralCompiler(dialect, statement)


def query_to_str(statement, bind=None):
    """
    returns a string of a sqlalchemy.orm.Query with parameters bound

    WARNING: this is dangerous and ONLY for testing, executing the results
    of this function can result in an SQL Injection attack.
    """
    if isinstance(statement, sqlalchemy.orm.Query):
        if bind is None:
            bind = statement.session.get_bind()
        statement = statement.statement
    elif bind is None:
        bind = statement.bind

    if bind is None:
        raise Exception(
            'bind param (engine or connection object) required when using with an unbound statement',  # noqa: E501
        )

    dialect = bind.dialect
    compiler = statement._compiler(dialect)
    literal_compiler = compiler_instance_factory(compiler, dialect, statement)
    return 'TESTING ONLY BIND: ' + literal_compiler.process(statement)


def assert_in_query(obj, test_for):
    query = obj.build_query() if hasattr(obj, 'build_query') else obj
    query_str = query_to_str(query)
    assert test_for in query_str, query_str


def assert_not_in_query(obj, test_for):
    query = obj.build_query() if hasattr(obj, 'build_query') else obj
    query_str = query_to_str(query)
    assert test_for not in query_str, query_str


def assert_list_equal(list1, list2):
    """
    A list-specific equality assertion.

    This method is based on the Python `unittest.TestCase.assertListEqual` method.

    :param list1:
    :param list2:
    :return:
    """

    # resolve generators
    list1, list2 = map(list, (list1, list2))

    assert len(list1) == len(list2), f'Lists are different lengths: {len(list1)} != {len(list2)}'

    if list1 == list2:
        # the lists are the same, we're done
        return

    # the lists are different in at least one element; find it
    # and report it
    for index, (val1, val2) in enumerate(zip(list1, list2, strict=True)):
        assert val1 == val2, f'First differing element at index {index}: {val1!r} != {val2!r}'


def assert_rendered_xlsx_matches(rendered_xlsx, xlsx_headers, xlsx_rows):
    """
    Verifies that `rendered_xlsx` has a set of headers and values that match
    the given parameters.

    NOTE: This method does not perform in-depth analysis of complex workbooks!
          Assumes header rows and data rows are contiguous.
          Multiple worksheets or complex layouts *are not verified!*

    :param rendered_xlsx: binary data passed to openpyxl as file contents
    :param xlsx_headers: list of rows of column headers
    :param xlsx_rows: list of rows in order as they will appear in the worksheet
    """
    assert rendered_xlsx
    rendered_xlsx.filename.seek(0)

    if not openpyxl:
        raise Exception('openpyxl is required for webgrid testing helpers to read XLSX')

    book = openpyxl.load_workbook(rendered_xlsx.filename)
    assert len(book.sheetnames) >= 1
    sheet = book[book.sheetnames[0]]

    # # verify the shape of the sheet

    # ## shape of rows (1 row for the headers, 1 for each row of data)
    nrows = len(xlsx_rows)
    if xlsx_headers:
        nrows += len(xlsx_headers)
    assert max([nrows, 1]) == sheet.max_row, (
        f'Sheet max row mismatch, {max([nrows, 1])} != {sheet.max_row}'
    )

    # ## shape of columns
    ncols = max(
        max(len(values) for values in xlsx_headers) if xlsx_headers else 0,
        max(len(values) for values in xlsx_rows) if xlsx_rows else 0,
    )
    assert max([ncols, 1]) == sheet.max_column, (
        f'Sheet max column mismatch, {max([ncols, 1])} != {sheet.max_column}'
    )

    row_iter = sheet.iter_rows()

    expected_rows = (xlsx_headers or []) + (xlsx_rows or [])

    for row, expected_row in zip(row_iter, expected_rows, strict=True):
        assert_list_equal((cell.value for cell in row), expected_row)


class GridBase:
    """Base test class for Flask or Keg apps.

    Class Attributes:
        grid_cls: Application grid class to use during testing

        filters: Iterable of (name, op, value, expected) tuples to check for filter logic,
        or a callable returning such an iterable. `name` is the column key. `op` and `value`
        set the filter parameters. `expected` is either a SQL string or compiled regex to
        find when the filter is enabled.

        sort_tests: Iterable of (name, expected) tuples to check for sort logic. `name` is
        the column key. `expected` is a SQL string to find when the sort is enabled.
    """

    grid_cls = None
    filters = ()
    sort_tests = ()

    @classmethod
    def setup_class(cls):
        if hasattr(cls, 'init'):
            cls.init()

    def query_to_str(self, statement, bind=None):
        """Render a SQLAlchemy query to a string."""
        return query_to_str(statement, bind=bind)

    def assert_in_query(self, look_for, grid=None, _query_string=None, **kwargs):
        """Verify the given SQL string is in the grid's query.

        Args:
            look_for (str): SQL string to find.

            grid (BaseGrid, optional): Grid to use instead of `self.get_session_grid`.
            Defaults to None.

            kwargs (dict, optional): Additional args passed to `self.get_session_grid`.
        """
        grid = grid or self.get_session_grid(_query_string=_query_string, **kwargs)
        assert_in_query(grid, look_for)

    def assert_not_in_query(self, look_for, grid=None, _query_string=None, **kwargs):
        """Verify the given SQL string is not in the grid's query.

        Args:
            look_for (str): SQL string to find.

            grid (BaseGrid, optional): Grid to use instead of `self.get_session_grid`.
            Defaults to None.

            kwargs (dict, optional): Additional args passed to `self.get_session_grid`.
        """
        grid = grid or self.get_session_grid(_query_string=_query_string, **kwargs)
        assert_not_in_query(grid, look_for)

    def assert_regex_in_query(self, look_for, grid=None, _query_string=None, **kwargs):
        """Verify the given regex matches the grid's query.

        Args:
            look_for (str or regex): Regex to search (can be compiled or provided as string).

            grid (BaseGrid, optional): Grid to use instead of `self.get_session_grid`.
            Defaults to None.

            kwargs (dict, optional): Additional args passed to `self.get_session_grid`.
        """
        grid = grid or self.get_session_grid(_query_string=_query_string, **kwargs)
        query_str = self.query_to_str(grid.build_query())

        if hasattr(look_for, 'search'):
            assert look_for.search(query_str), f'"{look_for.pattern}" not found in: {query_str}'
        else:
            assert re.search(look_for, query_str), f'"{look_for}" not found in: {query_str}'

    def get_grid(self, grid_args, *args, **kwargs):
        """Construct grid from args and kwargs, and apply grid_args.

        Args:
            grid_args: grid query args

        Returns:
            grid instance
        """
        grid = self.grid_cls(*args, **kwargs)
        grid.apply_qs_args(add_user_warnings=False, grid_args=grid_args)
        return grid

    def get_session_grid(self, *args, _query_string=None, **kwargs):
        """Construct grid from args and kwargs, and apply query string.

        Args:
            _query_string: URL query string with grid query args

        Returns:
            grid instance
        """
        grid = self.grid_cls(*args, **kwargs)
        if grid.manager.request():
            # request context already exists
            grid.apply_qs_args()
        else:
            url = f'/?{_query_string}' if _query_string else '/'
            with grid.manager.test_request_context(url=url):
                grid.apply_qs_args()
        return grid

    def get_pyq(self, grid=None, _query_string=None, **kwargs):
        """Turn provided/constructed grid into a rendered PyQuery object.

        Args:
            grid (BaseGrid, optional): Grid to use instead of `self.get_session_grid`.
            Defaults to None.

            kwargs (dict, optional): Additional args passed to `self.get_session_grid`.

        Returns:
            PyQuery object
        """
        session_grid = grid or self.get_session_grid(_query_string=_query_string, **kwargs)
        if session_grid.manager.request():
            # request context already exists
            html = session_grid.html()
        else:
            url = f'/?{_query_string}' if _query_string else '/'
            with session_grid.manager.test_request_context(url=url):
                html = session_grid.html()
        return PyQuery(f'<html>{html}</html>')

    def check_filter(self, name, op, value, expected):
        """Assertions to perform on a filter test.

        Args:
            name (str): Column key to filter.
            op (str): Filter operator to enable.
            value (Any): Filter value to assign.
            expected (str or regex): SQL string or compiled regex to find.
        """
        qs_args = [(f'op({name})', op)]
        if isinstance(value, (list, tuple)):
            for v in value:
                qs_args.append((f'v1({name})', v))
        else:
            qs_args.append((f'v1({name})', value))

        def sub_func(ex):
            query_string = urllib.parse.urlencode(qs_args)
            if isinstance(ex, re.compile('').__class__):
                self.assert_regex_in_query(ex, _query_string=query_string)
            else:
                self.assert_in_query(ex, _query_string=query_string)
            # ensures the query executes and the grid renders without error
            self.get_pyq(_query_string=query_string)

        def page_func():
            query_string = urllib.parse.urlencode([('onpage', 2), ('perpage', 1), *qs_args])
            pg = self.get_session_grid(_query_string=query_string)
            if pg.page_count > 1:
                self.get_pyq(_query_string=query_string)

        if self.grid_cls.pager_on:
            page_func()

        return sub_func(expected)

    def test_filters(self):
        """Use filters attribute/property/method to run assertions."""
        cases = self.filters() if callable(self.filters) else self.filters
        for name, op, value, expected in cases:
            self.check_filter(name, op, value, expected)

    def check_sort(self, k, ex, asc):
        """Assertions to perform on a sort test.

        Args:
            k (str): Column key to sort.
            ex (str or regex): SQL string to find.
            asc (bool): Flag indicating ascending/descending order.
        """
        if not asc:
            k = '-' + k
        d = {'sort1': k}

        def sub_func():
            query_string = urllib.parse.urlencode(d)
            self.assert_in_query(
                'ORDER BY {}{}'.format(ex, '' if asc else ' DESC'),
                _query_string=query_string,
            )
            # ensures the query executes and the grid renders without error
            self.get_pyq(_query_string=query_string)

        return sub_func()

    def test_sort(self):
        """Use sort_tests attribute/property to run assertions."""
        for col, expect in self.sort_tests:
            self.check_sort(col, expect, True)
            self.check_sort(col, expect, False)

    def _compare_table_block(self, block_selector, tag, expect):
        print(block_selector)
        assert len(block_selector) == len(expect)

        for row_idx, row in enumerate(expect):
            cells = block_selector.eq(row_idx).find(tag)
            assert len(cells) == len(row)
            for col_idx, val in enumerate(row):
                read = cells.eq(col_idx).text()
                assert read == val, f'row {row_idx} col {col_idx} {read} != {val}'

    def expect_table_header(self, expect, grid=None, _query_string=None, **kwargs):
        """Run assertions to compare rendered headings with expected data.

        Args:
            expect (list): List representation of expected table data.

            grid (BaseGrid, optional): Grid to use instead of `self.get_session_grid`.
            Defaults to None.

            kwargs (dict, optional): Additional args passed to `self.get_session_grid`.
        """
        d = self.get_pyq(grid, _query_string=_query_string, **kwargs)
        self._compare_table_block(
            d.find('table.records thead tr'),
            'th',
            expect,
        )

    def expect_table_contents(self, expect, grid=None, _query_string=None, **kwargs):
        """Run assertions to compare rendered data rows with expected data.

        Args:
            expect (list): List representation of expected table data.

            grid (BaseGrid, optional): Grid to use instead of `self.get_session_grid`.
            Defaults to None.

            kwargs (dict, optional): Additional args passed to `self.get_session_grid`.
        """
        d = self.get_pyq(grid, _query_string=_query_string, **kwargs)
        self._compare_table_block(
            d.find('table.records tbody tr'),
            'td',
            expect,
        )

    def test_search_expr_passes(self, grid=None, _query_string=None):
        """Assert that a single-search query executes without error."""
        grid = grid or self.get_session_grid(_query_string=_query_string)
        if grid.enable_search:
            grid.records  # noqa: B018


class MSSQLGridBase(GridBase):
    """MSSQL dialect produces some string oddities compared to other dialects, such as
    having the N'foo' syntax for unicode strings instead of 'foo'. This can clutter
    tests a bit. Using MSSQLGridBase will patch that into the asserts, so that
    look_for will match whether it has the N-prefix or not.
    """

    def query_to_str_replace_type(self, compiled_query):
        """Same as query_to_str, but accounts for pyodbc type-specific rendering."""
        query_str = self.query_to_str(compiled_query)
        # pyodbc rendering includes an additional character for some strings,
        # like N'foo' instead of 'foo'. This is not relevant to what we're testing.
        return re.sub(r"(\(|WHEN|LIKE|ELSE|THEN|[,=\+])( ?)N'(.*?)'", r"\1\2'\3'", query_str)

    def assert_in_query(self, look_for, grid=None, context=None, _query_string=None, **kwargs):
        session_grid = grid or self.get_session_grid(_query_string=_query_string, **kwargs)
        query_str = self.query_to_str(session_grid.build_query())
        query_str_repl = self.query_to_str_replace_type(session_grid.build_query())
        assert look_for in query_str or look_for in query_str_repl, (
            f'"{look_for}" not found in: {query_str}'
        )

    def assert_not_in_query(self, look_for, grid=None, context=None, _query_string=None, **kwargs):
        session_grid = grid or self.get_session_grid(_query_string=_query_string, **kwargs)
        query_str = self.query_to_str(session_grid.build_query())
        query_str_repl = self.query_to_str_replace_type(session_grid.build_query())
        assert look_for not in query_str or look_for not in query_str_repl, (
            f'"{look_for}" found in: {query_str}'
        )

    def assert_regex_in_query(
        self,
        look_for,
        grid=None,
        context=None,
        _query_string=None,
        **kwargs,
    ):
        session_grid = grid or self.get_session_grid(_query_string=_query_string, **kwargs)
        query_str = self.query_to_str(session_grid.build_query())
        query_str_repl = self.query_to_str_replace_type(session_grid.build_query())

        if hasattr(look_for, 'search'):
            assert look_for.search(query_str) or look_for.search(query_str_repl), (
                f'"{look_for.pattern}" not found in: {query_str}'
            )
        else:
            assert re.search(look_for, query_str) or re.search(look_for, query_str_repl), (
                f'"{look_for}" not found in: {query_str}'
            )
